import solara
from solara.website.components.sidebar import Sidebar
from solara.website.components.contact import Contact

Sidebar = Sidebar


@solara.component
def Page():
    solara.Markdown("""
# Contact

## Business

Solara is developed mostly by Maarten Breddels and Mario Buikhuizen, Founders
of [Widgetti](https://widgetti.io/).

For business inquiries, please reach out using the form below, or send an email to [contact@solara.dev](mailto:contact@solara.dev) regarding Solara, or [info@widgetti.io](mailto:info@widgetti.io) regarding more general Jupyter related matters.
""")
    Contact(
        title="Contact Us", subtitle="Please fill out the form below and we will get back to you as soon as possible.", email_subject="Contact Form Submission"
    )
    solara.Markdown(
        """
## Community

If you want to share your thoughts, share your experiences, or just want to talk
to other people using and developing on Solara, consider using [GitHub discussions](https://github.com/widgetti/solara/discussions) for asynchronous discussions
or consider joining discord for more synchronous discussions.


[![Discord Shield](https://discordapp.com/api/guilds/1106593685241614489/widget.png?style=banner2)](https://discord.gg/dm4GKNDjXN)
"""
    )



from typing import cast, Optional
import httpx
import asyncio
import solara
import solara.lab


@solara.component
def Page():
    btc = solara.use_reactive(cast(Optional[float], None))

    async def fetch_btc_price():
        while True:
            await asyncio.sleep(1)
            async with httpx.AsyncClient() as client:
                url = "https://api.binance.com/api/v1/ticker/price?symbol=BTCUSDT"
                response = await client.get(url)
                btc.value = float(response.json()["price"])
                print("btc.value", btc.value)

    fetch_result = solara.lab.use_task(fetch_btc_price, dependencies=[])
    # the task keeps running, so is always in the pending mode, so we combine it with the btc value being None
    if fetch_result.pending and btc.value is None:
        solara.Text("Fetching BTC price...")
    else:
        if fetch_result.error:
            solara.Error(f"Error fetching BTC price: {fetch_result.exception}")
        else:
            solara.Text(f"BTC price: ${btc.value}")


Page()



import contextlib
import logging
import os
from pathlib import Path
import pdb
import sys
import traceback

from rich import print

logger = logging.getLogger("solara.server")


def start_error(title, msg, exception: Exception = None):
    if exception:
        traceback.print_exception(None, exception, exception.__traceback__)
    print(f"[red]{title}:\n\t[blue]{msg}")  # noqa
    os._exit(-1)


def path_is_child_of(path: Path, parent: Path) -> bool:
    # We use os.path.normpath() because we do not want to follow symlinks
    # in editable installs, since some packages are symlinked
    path_string = os.path.normpath(path)
    parent_string = os.path.normpath(parent)
    if sys.platform == "win32":
        # on windows, we sometimes get different casing (only seen on CI)
        path_string = path_string.lower()
        parent_string = parent_string.lower()
    return path_string.startswith(parent_string)


@contextlib.contextmanager
def pdb_guard():
    from . import settings

    try:
        yield
    except Exception:
        if settings.main.use_pdb:
            logger.exception("Exception, will be handled by debugger")
            pdb.post_mortem()
        else:
            raise



"""
# use_state_or_update

Like `solara.use_state`, except it will update the value if the input value changes.

This is useful to give a component state *or* allow the state to be controlled by the parent component


"""

import solara

title = "use_state_or_update"


@solara.component
def SliderWithoutState(value: int):
    # Note that this is a very bad practice, if value is an input change and the slider
    # can change it, this component should have a on_value callback to allow the parent
    # component to manage its state.
    # This component is only for demoing/understanding.
    return solara.IntSlider("value", value=value)


@solara.component
def SliderWithState(value: int):
    # effectively, because of use_state, the value prop passed in is
    # only a default value.
    value, set_value = solara.use_state(value)
    return solara.IntSlider("value", value=value, on_value=set_value)


@solara.component
def SliderWithStateOrUpdate(value: int):
    value, set_value = solara.use_state_or_update(value)
    return solara.IntSlider("value", value=value, on_value=set_value)


@solara.component
def Page():
    parent_value, set_parent_value = solara.use_state(4)
    # used to force rerenders
    rerender_counter, set_rerender_counter = solara.use_state(4)
    with solara.Card("Parent value selection"):
        solara.Info("This slider value gets passed down to the child components")
        solara.IntSlider("parent value", value=parent_value, on_value=set_parent_value)
        solara.Button("Force redraw", on_click=lambda: set_rerender_counter(rerender_counter + 1))

    with solara.Card("Child without state"):
        solara.Info("This child will simply render the value passed into the argument, a redraw will reset it to its parent value.")
        SliderWithoutState(parent_value)

    with solara.Card("Child with state"):
        solara.Info("This child will not care about the value passed into the prop, it manages its own state.")
        SliderWithState(parent_value)

    with solara.Card("Child with state (or update)"):
        solara.Info("This child will update when the passes in a new value, but a redraw will not reset it.")
        SliderWithStateOrUpdate(parent_value)

    with solara.Card("Child with state + key"):
        solara.Info(
            "We can also use the `.key(...)` method to force the component to forget its state, this will however cause the widget to be re-created"
            "(a performance penalty)."
        )
        SliderWithState(parent_value).key(f"slider-{parent_value}")



from .chat import ChatBox, ChatInput, ChatMessage  # noqa: F401
from .confirmation_dialog import ConfirmationDialog  # noqa: F401
from .input_date import InputDate, InputDateRange  # noqa: F401
from .input_time import InputTime as InputTime
from .menu import ClickMenu, ContextMenu, Menu  # noqa: F401 F403
from .tabs import Tab, Tabs  # noqa: F401
from .theming import ThemeToggle, theme, use_dark_effective  # noqa: F401



import functools
import inspect
import logging
import os
import threading
from typing import Callable, Iterator, Optional, TypeVar, Union, cast

import solara
from solara.datatypes import Result, ResultState
from solara.util import cancel_guard, nullcontext

SOLARA_ALLOW_OTHER_TRACER = os.environ.get("SOLARA_ALLOW_OTHER_TRACER", False) in (True, "True", "true", "1")
T = TypeVar("T")
logger = logging.getLogger("solara.hooks.use_thread")


def use_thread(
    callback: Union[
        Callable[[threading.Event], T],
        Iterator[Callable[[threading.Event], T]],
        Callable[[], T],
        Iterator[Callable[[], T]],
    ],
    dependencies=[],
    intrusive_cancel=True,
) -> Result[T]:
    from .misc import use_force_update, use_retry

    def make_event(*_ignore_dependencies):
        return threading.Event()

    def make_lock():
        return threading.Lock()

    lock: threading.Lock = solara.use_memo(make_lock, [])
    updater = use_force_update()
    result_state, set_result_state = solara.use_state(ResultState.INITIAL)
    error = solara.use_ref(cast(Optional[Exception], None))
    result = solara.use_ref(cast(Optional[T], None))
    running_thread = solara.use_ref(cast(Optional[threading.Thread], None))
    counter, retry = use_retry()
    cancel: threading.Event = solara.use_memo(make_event, [*dependencies, counter])

    def run():
        set_result_state(ResultState.STARTING)

        def runner():
            wait_for_thread = None
            with lock:
                # if there is a current thread already, we'll need
                # to wait for it. copy the ref, and set ourselves
                # as the current one
                if running_thread.current:
                    wait_for_thread = running_thread.current
                running_thread.current = threading.current_thread()
            if wait_for_thread is not None:
                set_result_state(ResultState.WAITING)
                # don't start before the previous is stopped
                try:
                    wait_for_thread.join()
                except:  # noqa
                    pass
                if threading.current_thread() != running_thread.current:
                    # in case a new thread was started that also was waiting for the previous
                    # thread to st stop, we can finish this
                    return
            # we previously set current to None, but if we do not do that, we can still render the old value
            # while we can still show a loading indicator using the .state
            # result.current = None
            set_result_state(ResultState.RUNNING)

            sig = inspect.signature(callback)  # type: ignore
            if sig.parameters:
                f = functools.partial(callback, cancel)  # type: ignore
            else:
                f = callback  # type: ignore
            try:
                try:
                    # we only use the cancel_guard context manager around
                    # the function calls to f. We don't want to guard around
                    # a call to react, since that might slow down rendering
                    # during rendering
                    with cancel_guard(cancel) if intrusive_cancel else nullcontext():
                        value = f()
                    if inspect.isgenerator(value):
                        while True:
                            try:
                                with cancel_guard(cancel) if intrusive_cancel else nullcontext():
                                    result.current = next(value)
                                    error.current = None
                            except StopIteration:
                                break
                            # assigning to the ref doesn't trigger a rerender, so do it manually
                            updater()
                        if threading.current_thread() == running_thread.current:
                            set_result_state(ResultState.FINISHED)
                    else:
                        result.current = value
                        error.current = None
                        if threading.current_thread() == running_thread.current:
                            set_result_state(ResultState.FINISHED)
                except Exception as e:
                    error.current = e
                    if threading.current_thread() == running_thread.current:
                        logger.exception(e)
                        set_result_state(ResultState.ERROR)
                    return
                except solara.util.CancelledError:
                    pass
                    # this means this thread is cancelled not be request, but because
                    # a new thread is running, we can ignore this
            finally:
                if threading.current_thread() == running_thread.current:
                    running_thread.current = None
                    logger.info("thread done!")
                    if cancel.is_set():
                        set_result_state(ResultState.CANCELLED)

        logger.info("starting thread: %r", runner)
        thread = threading.Thread(target=runner, daemon=True)
        thread.start()

        def cleanup():
            cancel.set()  # cleanup for use effect

        return cleanup

    solara.use_effect(run, dependencies + [counter])
    return Result[T](value=result.current, error=error.current, state=result_state, cancel=cancel.set, _retry=retry)



"""
# Tab

"""

import solara
import solara.lab
from solara.website.utils import apidoc

disabled = solara.reactive(False)
icon = solara.reactive(True)


@solara.component
def Page():
    solara.Checkbox(label="Disable Tab 2", value=disabled)
    solara.Checkbox(label="Show icon", value=icon)
    with solara.lab.Tabs():
        with solara.lab.Tab("Tab 1"):
            solara.Markdown("Hello")
        with solara.lab.Tab("Tab 2", disabled=disabled.value, icon_name="mdi-home" if icon.value else None):
            solara.Markdown("World")


__doc__ += apidoc(solara.lab.Tab.f)  # type: ignore



import dataclasses
import math
import os
from dataclasses import replace
from typing import Any, Callable, List, Optional, cast

import ipyvuetify as v
import ipywidgets
import solara
import solara.hooks.dataframe
import solara.lab
import traitlets
from solara.lab.hooks.dataframe import use_df_column_names, df_row_names
from solara.lab.utils.dataframe import df_len, df_records, df_slice

from .. import CellAction, ColumnAction


def _ensure_dict(d):
    if dataclasses.is_dataclass(d) and not isinstance(d, type):  # is_dataclass also returns True for dataclass type, rather than instance
        return dataclasses.asdict(d)
    return d


def _drop_keys_from_list_of_mappings(drop):
    def closure(list_of_dicts, widget):
        return [{k: v for k, v in _ensure_dict(d).items() if k not in drop} for d in list_of_dicts]

    return closure


class DataTableWidget(v.VuetifyTemplate):
    template_file = os.path.realpath(os.path.join(os.path.dirname(__file__), "datatable.vue"))

    total_length = traitlets.CInt().tag(sync=True)
    checked = traitlets.List(cast(List[Any], [])).tag(sync=True)  # indices of which rows are selected
    column_actions = traitlets.List(trait=traitlets.Instance(ColumnAction), default_value=[]).tag(
        sync=True, to_json=_drop_keys_from_list_of_mappings(["on_click"])
    )
    _column_actions_callbacks = traitlets.List(trait=traitlets.Callable(), default_value=[])
    cell_actions = traitlets.List(trait=traitlets.Instance(CellAction), default_value=[]).tag(sync=True, to_json=_drop_keys_from_list_of_mappings(["on_click"]))
    _cell_actions_callbacks = traitlets.List(trait=traitlets.Callable(), default_value=[])
    items = traitlets.Any().tag(sync=True)  # the data, a list of dict
    headers = traitlets.Any().tag(sync=True)
    headers_selections = traitlets.Any().tag(sync=True)
    options = traitlets.Any().tag(sync=True)
    items_per_page = traitlets.CInt(11).tag(sync=True)
    selections = traitlets.Any([]).tag(sync=True)
    selection_colors = traitlets.Any([]).tag(sync=True)
    selection_enabled = traitlets.Bool(True).tag(sync=True)
    highlighted = traitlets.Int(None, allow_none=True).tag(sync=True)
    scrollable = traitlets.Bool(False).tag(sync=True)

    # for use with scrollable, when used in the default UI
    height = traitlets.Unicode(None, allow_none=True).tag(sync=True)

    hidden_components = traitlets.List(cast(List[Any], [])).tag(sync=False)
    column_header_hover = traitlets.Unicode(allow_none=True).tag(sync=True)
    column_header_widget = traitlets.Any(allow_none=True).tag(sync=True, **ipywidgets.widget_serialization)

    def vue_on_column_action(self, data):
        header_value, action_index = data
        on_click = self._column_actions_callbacks[action_index]
        if on_click:
            on_click(header_value)

    def vue_on_cell_action(self, data):
        row, header_value, action_index = data
        on_click = self._cell_actions_callbacks[action_index]
        if on_click:
            on_click(header_value, row)


def format_default(df, column, row_index, value):
    if isinstance(value, float) and math.isnan(value):
        return "NaN"
    return str(value)


@solara.component
def DataTable(
    df,
    page=0,
    items_per_page=20,
    format=None,
    column_actions: List[ColumnAction] = [],
    cell_actions: List[CellAction] = [],
    scrollable=False,
    on_column_header_hover: Optional[Callable[[Optional[str]], None]] = None,
    column_header_info: Optional[solara.Element] = None,
):
    total_length = df_len(df)
    options = {"descending": False, "page": page + 1, "itemsPerPage": items_per_page, "sortBy": [], "totalItems": total_length}
    options, set_options = solara.use_state(options, key="options")
    format = format or format_default
    # frontend does 1 base, we use 0 based
    page = options["page"] - 1
    items_per_page = options["itemsPerPage"]
    i1 = page * items_per_page
    i2 = min(total_length, (page + 1) * items_per_page)

    columns = use_df_column_names(df)
    rows = df_row_names(df)
    items = []
    dfs = df_slice(df, i1, i2)
    records = df_records(dfs)
    for i in range(i2 - i1):
        item = {"__row__": format(dfs, columns, i + 1, rows[i + i1])}  # special key for the row number
        for column in columns:
            item[column] = format(dfs, column, i + i1, records[i][column])
        items.append(item)

    headers = [{"text": name, "value": name, "sortable": False} for name in columns]
    column_actions_callbacks = [k.on_click for k in column_actions]
    cell_actions_callbacks = [k.on_click for k in cell_actions]
    column_actions = [replace(k, on_click=None) for k in column_actions]
    cell_actions = [replace(k, on_click=None) for k in cell_actions]

    return DataTableWidget.element(
        total_length=total_length,
        items=items,
        headers=headers,
        headers_selections=[],
        options=options,
        items_per_page=items_per_page,
        selections=[],
        selection_colors=[],
        selection_enabled=False,
        highlighted=None,
        scrollable=scrollable,
        on_options=set_options,
        column_actions=column_actions,
        cell_actions=cell_actions,
        _column_actions_callbacks=column_actions_callbacks,
        _cell_actions_callbacks=cell_actions_callbacks,
        on_column_header_hover=on_column_header_hover,
        column_header_widget=column_header_info,
    )


@solara.component
def DataFrame(
    df,
    items_per_page=20,
    column_actions: List[ColumnAction] = [],
    cell_actions: List[CellAction] = [],
    scrollable=False,
    on_column_header_hover: Optional[Callable[[Optional[str]], None]] = None,
    column_header_info: Optional[solara.Element] = None,
):
    """Displays a Pandas dataframe in a table.

    Pass in a dataframe as first argument, and optionally how many items per page to display.

    ```solara
    import solara
    import pandas as pd
    import plotly

    df = plotly.data.iris()

    @solara.component
    def Page():
        solara.DataFrame(df, items_per_page=5)

    ```

    # Custom column header info

    Use the `column_header_info` argument to display a custom component on the column header when
    the user hover above it. In this case we display the value counts for the column.

    ```solara
    import solara
    import pandas as pd
    import plotly

    df = plotly.data.iris()

    @solara.component
    def Page():
        column_hover, set_column_hover = solara.use_state(None)

        with solara.Column(margin=4) as column_header_info:
            if column_hover:
                solara.Text("Value counts for " + column_hover)
                display(df[column_hover].value_counts())
            # if no column is hovered above, we provide an empty container
            # so we always see the triple dot icon on the column header

        solara.DataFrame(df, column_header_info=column_header_info, on_column_header_hover=set_column_hover)
    ```


    ## Arguments

     * `df` - `DataFrame` - a Pandas dataframe.
     * `items_per_page` - `int` - number of items per page.
     * `column_actions` - Triggered via clicking on the triple dot icon on the headers (visible when hovering).
     * `cell_actions` -  Triggered via clicking on the triple dot icon in the cell (visible when hovering).
     * `on_column_header_hover` - Optional callback when the user hovers over the triple dot icon on a header.
     * `column_header_info` - Element to display in the column menu popup (visible when hovering), provide an
            empty container element (like [Column](/documentation/components/layout/column)) to force showing the trigle dot icon (see example).

    """
    return DataTable(
        df,
        items_per_page=items_per_page,
        column_actions=column_actions,
        cell_actions=cell_actions,
        scrollable=scrollable,
        on_column_header_hover=on_column_header_hover,
        column_header_info=column_header_info,
    )



import solara


@solara.component
def Page():
    with solara.Column():
        solara.Markdown(
            """
            # Careers
            We are always looking for talented individuals to join our team. If you're still be interested after reading the rest of this page, reach out at [contact@widgetti.io](mailto:contact@widgetti.io).

            ## Who Are We?
            We are a small team of developers working in the gap between Jupyter notebooks and web applications. We are passionate about making data science more accessible and app development painless.
            We are based in Groningen, The Netherlands, but work remote first.

            ## What Do We Offer?
            We offer a flexible work environment, competitive salary, and the opportunity to do meaningful, crossdisciplinary work with scientists and companies worldwide.
            We are a small team, so you will have a lot of freedom in your work, and the chance to get your hands dirty with many projects.

            ### Open Source
            We are strong believers in open source software. When working with us, you can expect the majority of the software you write to be open source, with contributions to Solara as well as other open source projects.

            ## Who Are We Looking For?
            We are looking for developers, technical writers, designers, and everything in between.
            """,
            style={"max-width": "90%", "margin": "0 auto"},
        )



import inspect
import os
from pathlib import Path
from typing import Any, Optional, List, Type

# similar API to pydantic/pydantic-settings but we prefer not to have a dependency on pydantic
# since we cannot be compatible with pydantic1 and 2
# NOTE: not a public api


def _get_type(annotation):
    check_optional_types = [str, int, float, bool, dict, list]
    for check_type in check_optional_types:
        if annotation == Optional[check_type]:
            return check_type
    if hasattr(annotation, "__origin__"):
        if annotation.__origin__ is dict:
            return dict
    return annotation


class _Field:
    def __init__(self, default=None, env=None, title=None, default_factory=None, gt=None, alias=None) -> None:
        self.default = default
        self.env = env
        self.fullenv = None
        self.title = title
        self.annotation = None
        self.default_factory = default_factory
        self.gt = gt
        self.alias = alias
        self.field_info = self
        self.extra = {"env_names": [env] if env else []}

    def __set_name__(self, owner, name):
        prefix = "SOLARA_"
        config = getattr(owner, "Config")
        if config:
            prefix = getattr(config, "env_prefix", prefix).upper()
            if hasattr(config, "fields"):
                fields = config.fields
                if name in fields:
                    self.alias = fields[name]
        self.name = name
        self.alias = self.alias or self.name
        self.title = self.title or self.name
        if self.env is None:
            self.env = f"{prefix}{self.name.upper()}"
        else:
            self.env = self.env
        self.annotation = owner.__annotations__.get(self.name)
        assert self.annotation is not None, f"Field {self.name} must have a type annotation"
        self.type_ = _get_type(self.annotation)

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return instance._values[self.name]

    def __set__(self, instance, value):
        instance._values[self.name] = value


def convert(annotation, value: str) -> Any:
    check_sub_types: List[Type] = [str, int, float, bool, Path]
    for sub_type in check_sub_types:
        if annotation == Optional[sub_type]:
            annotation = sub_type
            return convert(annotation, value)
    for sub_type in check_sub_types:
        if annotation == List[sub_type]:  # type: ignore
            annotation = sub_type
            values = value.split(",")
            return [convert(sub_type, k) for k in values]
    if annotation is str:
        return value
    elif annotation is int:
        return int(value)
    elif annotation is float:
        return float(value)
    elif annotation is bool:
        if value in ("True", "true", "1"):
            return True
        elif value in ("False", "false", "0"):
            return False
        else:
            raise ValueError(f"Invalid boolean value {value}")
    else:
        # raise TypeError(f"Unsupported type {annotation}")
        return annotation(value)


def Field(*args, **kwargs) -> Any:
    return _Field(*args, **kwargs)


class BaseSettings:
    __fields__: dict

    def __init__(self, **kwargs) -> None:
        cls = type(self)
        self._values = {**kwargs}
        keys = {k.upper() for k in os.environ.keys()}
        for key, field in cls.__dict__.items():
            if key in kwargs:
                continue
            if isinstance(field, _Field):
                value = field.default
                if field.default_factory:
                    value = field.default_factory()

                if field.env:
                    env_key = field.env.upper()
                    if env_key in keys:
                        # do a case-insensitive lookup
                        for env_var_cased in os.environ.keys():
                            if env_key.upper() == env_var_cased.upper():
                                value = convert(field.annotation, os.environ[env_var_cased])
                self._values[key] = value

    def __init_subclass__(cls) -> None:
        cls.__fields__ = {}
        for key, field in cls.__dict__.items():
            if key.startswith("_"):
                continue
            if key == "Config":
                continue
            if not isinstance(field, _Field):
                if inspect.isfunction(field):
                    continue
                field = Field(field)
                setattr(cls, key, field)
                field.__set_name__(cls, key)
            cls.__fields__[key] = field

    def dict(self, by_alias=True):
        values = self._values.copy()
        for key, value in values.items():
            if isinstance(value, BaseSettings):
                values[key] = value.dict(by_alias=by_alias)
        return values



"""
# InputTime
"""

import solara
from solara.website.components import NoPage
from solara.website.utils import apidoc

title = "InputTime"


__doc__ += apidoc(solara.lab.components.input_time.InputTime.f)  # type: ignore


Page = NoPage



import hashlib
import html
import logging
import textwrap
import traceback
import warnings
from typing import Any, Callable, Dict, List, Optional, Union, cast
import typing

import ipyvuetify as v

try:
    import pymdownx.emoji
    import pymdownx.highlight
    import pymdownx.superfences

    has_pymdownx = True
except ModuleNotFoundError:
    has_pymdownx = False
import reacton.core

import solara
import solara.components.applayout

try:
    import pygments

    has_pygments = True
except ModuleNotFoundError:
    has_pygments = False
else:
    from pygments.formatters import HtmlFormatter
    from pygments.lexers import get_lexer_by_name

if typing.TYPE_CHECKING:
    import markdown

logger = logging.getLogger(__name__)

html_no_execute_enabled = "<div><i>Solara execution is not enabled</i></div>"


@solara.component
def ExceptionGuard(children=[]):
    exception, clear_exception = solara.use_exception()
    if exception:
        solara.Error(f"Oops, an error occurred: {str(exception)}")
        with solara.Details("Exception details"):
            error = "".join(traceback.format_exception(None, exception, exception.__traceback__))
            solara.Preformatted(error)
    else:
        if len(children) == 1:
            return children[0]
        else:
            solara.Column(children=children)


def _run_solara(code, cleanups):
    ast = compile(code, "markdown", "exec")
    local_scope: Dict[Any, Any] = {}
    exec(ast, local_scope)
    app = None
    if "app" in local_scope:
        app = local_scope["app"]
    elif "Page" in local_scope:
        Page = local_scope["Page"]
        app = solara.components.applayout._AppLayoutEmbed(children=[ExceptionGuard(children=[Page()])])
    else:
        raise NameError("No Page or app defined")
    box = v.Html(tag="div")

    rc: reacton.core.RenderContext

    def cleanup():
        rc.close()

    cleanups.append(cleanup)
    box, rc = solara.render(cast(solara.Element, app), container=box)  # type: ignore
    widget_id = box._model_id
    return (
        '<div class="solara-markdown-output v-card v-sheet elevation-7">'
        f'<jupyter-widget widget="IPY_MODEL_{widget_id}">loading widget...</jupyter-widget>'
        '<div class="v-messages">Live output</div></div>'
    )


def _markdown_template(
    html,
    style="",
):
    cdn = None
    import solara.settings

    if not solara.settings.assets.proxy:
        cdn = solara.settings.assets.cdn

    template = (
        """
<template>
    <div class="solara-markdown rendered_html jp-RenderedHTMLCommon" style=\""""
        + style
        + """\">"""
        + html
        + r"""</div>
</template>

<script>
module.exports = {
    async mounted() {
        this.cdn = """
        + (rf"'{cdn}'" if cdn is not None else r"null")
        + r""";
        await this.loadRequire();
        this.mermaid = await this.loadMermaid();
        this.mermaid.init();
        this.latexSettings = {
                delimiters: [
                    {left: "$$", right: "$$", display: true},
                    {left: "$", right: "$", display: false},
                    {left: "\\[", right: "\\]", display: true},
                    {left: "\\(", right: "\\)", display: false}
                ]
            };
        if (window.renderMathInElement) {
            window.renderMathInElement(this.$el, this.latexSettings);
        } else if (window.MathJax && MathJax.Hub) {
            MathJax.Hub.Queue(['Typeset', MathJax.Hub, this.$el]);
        } else {
            window.renderMathInElement = await this.loadKatexExt();
            window.renderMathInElement(this.$el, this.latexSettings);
        }
        this.$el.querySelectorAll("a").forEach(a => this.setupRouter(a))
        window.md = this.$el
    },
    methods: {
        setupRouter(a) {
            let href = a.attributes['href'].value;
            if(href.startsWith("./")) {
                // TODO: should we really do this?
                href = location.pathname + href.substr(1);
                a.attributes['href'].href = href;
            }
            let authLink = href.startsWith("/_solara/auth/");
            if( (href.startsWith("./") || href.startsWith("/")) && !authLink) {
                a.onclick = e => {
                    console.log("clicked", href)
                    if(href.startsWith("./")) {
                        solara.router.push(href);
                    } else {
                        solara.router.push(href);
                    }
                    e.preventDefault()
                }
            } else if(href.startsWith("#")) {
                href = location.pathname + href;
                a.attributes['href'].value = href;
            } else {
                console.log("href", href, "is not a local link")
            }
        },
        async loadKatex() {
            require.config({
                map: {
                    '*': {
                        'katex': `${this.getCdn()}/katex@0.16.9/dist/katex.min.js`,
                    }
                }
            });
            const link = document.createElement('link');
            link.type = "text/css";
            link.rel = "stylesheet";
            link.href = `${this.getCdn()}/katex@0.16.9/dist/katex.min.css`;
            document.head.appendChild(link);
        },
        async loadKatexExt() {
            this.loadKatex();
            return (await this.import([`${this.getCdn()}/katex@0.16.9/dist/contrib/auto-render.min.js`]))[0]
        },
        async loadMermaid() {
            return (await this.import([`${this.getCdn()}/mermaid@10.8.0/dist/mermaid.min.js`]))[0]
        },
        import(dependencies) {
            return this.loadRequire().then(
                () => {
                    if (window.jupyterVue) {
                        // in jupyterlab, we take Vue from ipyvue/jupyterVue
                        define("vue", [], () => window.jupyterVue.Vue);
                    } else {
                        define("vue", ['jupyter-vue'], jupyterVue => jupyterVue.Vue);
                    }
                    return new Promise((resolve, reject) => {
                        requirejs(dependencies, (...modules) => resolve(modules));
                    })
                }
            );
        },
        loadRequire() {
            if (window.requirejs) {
                return Promise.resolve();
            }
            return new Promise((resolve, reject) => {
                const script = document.createElement('script');
                script.src = `${this.getCdn()}/requirejs@2.3.6/require.min.js`;
                script.onload = resolve;
                script.onerror = reject;
                document.head.appendChild(script);
            });
        },
        getJupyterBaseUrl() {
            // if base url is set, we use ./ for relative paths compared to the base url
            if (document.getElementsByTagName("base").length) {
                return "./";
            }
            const labConfigData = document.getElementById('jupyter-config-data');
            if (labConfigData) {
                /* lab and Voila */
                return JSON.parse(labConfigData.textContent).baseUrl;
            }
            let base = document.body.dataset.baseUrl || document.baseURI;
            if (!base.endsWith('/')) {
                base += '/';
            }
            return base
        },
        getCdn() {
            return this.cdn || (window.solara ? window.solara.cdn : `${this.getJupyterBaseUrl()}_solara/cdn`);
        }
    },
    updated() {
        // if the html gets update, re-run mermaid
        this.mermaid.init();

        if(window.MathJax && MathJax.Hub) {
            MathJax.Hub.Queue(['Typeset', MathJax.Hub, this.$el]);
        } else {
            window.renderMathInElement(this.$el, this.latexSettings);
        }
    }
}
</script>
    """
    )
    return template


def _highlight(src, language, class_name=None, options=None, md=None, unsafe_solara_execute=False, cleanups=None, **kwargs):
    """Highlight a block of code"""
    if not has_pygments:
        warnings.warn("Pygments is not installed, code highlighting will not work, use pip install pygments to install it.")
        src_safe = html.escape(src)
        return f"<pre><code>{src_safe}</code></pre>"

    run_src_with_solara = False
    if language == "solara":
        run_src_with_solara = True
        language = "python"

    lexer = get_lexer_by_name(language)
    formatter = HtmlFormatter()
    src_html = pygments.highlight(src, lexer, formatter)

    if run_src_with_solara:
        if unsafe_solara_execute:
            html_widget = _run_solara(src, cleanups)
            return src_html + html_widget
        else:
            return src_html + html_no_execute_enabled
    else:
        return src_html


def formatter(unsafe_solara_execute: bool, cleanups: List[Callable[[], None]]):
    def wrapper(*args, **kwargs):
        try:
            kwargs["unsafe_solara_execute"] = unsafe_solara_execute
            kwargs["cleanups"] = cleanups
            return _highlight(*args, **kwargs)
        except Exception as e:
            logger.exception("Error while highlighting code")
            raise e

    return wrapper


@solara.component
def MarkdownIt(md_text: str, highlight: List[int] = [], unsafe_solara_execute: bool = False):
    md_text = textwrap.dedent(md_text)

    from markdown_it import MarkdownIt as MarkdownItMod
    from mdit_py_plugins import container, deflist  # noqa: F401
    from mdit_py_plugins.footnote import footnote_plugin  # noqa: F401
    from mdit_py_plugins.front_matter import front_matter_plugin  # noqa: F401

    cleanups = solara.use_ref(cast(List[Callable[[], None]], []))

    def highlight_code(code, name, attrs):
        return _highlight(cleanups.current, code, name, unsafe_solara_execute, attrs)

    md = MarkdownItMod(
        "js-default",
        {
            "html": True,
            "typographer": True,
            "highlight": highlight_code,
        },
    )
    md = md.use(container.container_plugin, name="note")
    html = md.render(md_text)
    hash = hashlib.sha256((html + str(unsafe_solara_execute) + repr(highlight)).encode("utf-8")).hexdigest()

    def cleanup_wrapper():
        def cleanup():
            for cleanup in cleanups.current:
                cleanup()

        return cleanup

    solara.use_effect(cleanup_wrapper)
    return v.VuetifyTemplate.element(template=_markdown_template(html)).key(hash)


if has_pymdownx:
    _index = pymdownx.emoji.emojione(None, None)


def _no_deep_copy_emojione(options, md):
    return _index


@solara.component
def Markdown(md_text: str, unsafe_solara_execute=False, style: Union[str, Dict, None] = None, md_parser: Optional["markdown.Markdown"] = None):
    """Renders markdown text

    Renders markdown using https://python-markdown.github.io/

    Math rendering is done using Latex syntax, using https://katex.org/.

    ## Examples

    ### Basic

    ```solara
    import solara


    @solara.component
    def Page():
        return solara.Markdown(r'''
        # This is a title

        ## This is a subtitle
        This is a markdown text, **bold** and *italic* text is supported.

        ## Math
        Also, $x^2$ is rendered as math.

        Or multiline math:
        $$
        \\int_0^1 x^2 dx = \\frac{1}{3}
        $$

        ''')
    ```

    ## Arguments

     * `md_text`: The markdown text to render
     * `unsafe_solara_execute`: If True, code marked with language "solara" will be executed. This is potentially unsafe
        if the markdown text can come from user input and should only be used for trusted markdown.
     * `style`: A string or dict of css styles to apply to the rendered markdown.
     * `md_parser`: A markdown object to use for rendering. If not provided, a markdown object will be created.

    """
    import markdown

    md_text = textwrap.dedent(md_text)
    style = solara.util._flatten_style(style)
    cleanups = solara.use_ref(cast(List[Callable[[], None]], []))

    def make_markdown_object():
        if md_parser is not None:
            # we won't use the use_memo
            return None
        if has_pymdownx:
            return markdown.Markdown(  # type: ignore
                extensions=[
                    "pymdownx.highlight",
                    "pymdownx.superfences",
                    "pymdownx.emoji",
                    "toc",  # so we get anchors for h1 h2 etc
                    "tables",
                ],
                extension_configs={
                    "pymdownx.emoji": {
                        "emoji_index": _no_deep_copy_emojione,
                    },
                    "pymdownx.superfences": {
                        "custom_fences": [
                            {
                                "name": "mermaid",
                                "class": "mermaid",
                                "format": pymdownx.superfences.fence_div_format,
                            },
                            {
                                "name": "solara",
                                "class": "",
                                "format": formatter(unsafe_solara_execute, cleanups=cleanups.current),
                            },
                        ],
                    },
                },
            )
        else:
            logger.warning("Pymdownx not installed, using default markdown. For a better experience, install pymdownx.")
            return markdown.Markdown(  # type: ignore
                extensions=[
                    "fenced_code",
                    "codehilite",
                    "toc",
                    "tables",
                ],
            )

    md_self = solara.use_memo(make_markdown_object, dependencies=[unsafe_solara_execute])
    if md_parser is None:
        assert md_self is not None
        md_parser = md_self
    html = md_parser.convert(md_text)

    def cleanup_wrapper():
        def cleanup():
            for cleanup in cleanups.current:
                cleanup()

        return cleanup

    solara.use_effect(cleanup_wrapper, [])
    # if we update the template value, the whole vue tree will rerender (ipvue/ipyvuetify issue)
    # however, using the hash we simply generate a new widget each time
    hash = hashlib.sha256((html + str(unsafe_solara_execute)).encode("utf-8")).hexdigest()
    return v.VuetifyTemplate.element(template=_markdown_template(html, style)).key(hash)



import operator
from functools import reduce
from typing import Any, Callable, Dict, List, TypeVar

import solara.util
from solara.hooks.misc import use_force_update, use_unique_key

T = TypeVar("T")

__all__ = [
    "provide_cross_filter",
    "use_cross_filter",
]


class CrossFilterStore:
    def __init__(self) -> None:
        self.listeners: List[Callable] = []
        self.filters: Dict[Any, Dict[str, Any]] = {}

    def add(self, data_key, key, filter):
        data_filters = self.filters.setdefault(data_key, {})
        data_filters[key] = filter

    def use(self, data_key, key, eq=None):
        # we use this state to trigger update, we could do without
        updater = use_force_update()

        data_filters = self.filters.setdefault(data_key, {})
        filter, set_filter = solara.use_state(data_filters.get(key), eq=eq)

        def on_change():
            set_filter(data_filters.get(key))
            # even if we don't change our own filter, the other may change
            updater()

        def connect():
            self.listeners.append(on_change)
            # we need to force an extra render after the first render
            # to make sure we have the correct filter, since others components
            # may set a filter after we have rendered, *or* mounted
            on_change()

            def cleanup():
                self.listeners.remove(on_change)
                # also remove our filter, and notify the rest
                data_filters.pop(key, None)  # remove, ignoring key error
                for listener in self.listeners:
                    listener()

            return cleanup

        solara.use_effect(connect, [key])

        def setter(filter):
            data_filters[key] = filter
            for listener in self.listeners:
                listener()

        otherfilters = [filter for key_other, filter in data_filters.items() if key != key_other and filter is not None]
        return filter, otherfilters, setter


cross_filter_context = solara.create_context(CrossFilterStore())


def provide_cross_filter():
    # create it once
    cross_filter_object = solara.use_memo(CrossFilterStore, [])
    cross_filter_context.provide(cross_filter_object)
    return cross_filter_object


def use_cross_filter(data_key, name: str = "no-name", reducer: Callable[[T, T], T] = operator.and_, eq=solara.util.numpy_equals):
    """Provides cross filtering, all other filters are combined using the reducer.

    Cross filtering will collect a set of filters (from other components), and combine
    them into a single filter, that excludes the filter we set for the current component.
    This is often used in dashboards where a filter is defined in a visualization component,
    but only applied to all other components.

    The graph below shows what happens when component A and B set a filter, and C does not.

    ```mermaid
    graph TD;
        A--"filter A"-->B;
        B--"filter B"-->C;
        A--"filter A"-->C;
        B--"filter B"-->A;
    ```
    """
    key = use_unique_key(prefix=f"cross-filter-{name}-")
    cross_filter_store = solara.use_context(cross_filter_context)
    _own_filter, otherfilters, set_filter = cross_filter_store.use(data_key, key, eq=eq)
    if otherfilters:
        cross_filter = reduce(reducer, otherfilters[1:], otherfilters[0])
    else:
        cross_filter = None
    return cross_filter, set_filter



import asyncio
import sys
import time
from unittest.mock import Mock

import pytest

import solara.server.server
import solara.server.settings
from solara.server import kernel_context

on_windows = sys.platform == "win32"


@pytest.fixture
def short_cull_timeout():
    cull_timeout_previous = solara.server.settings.kernel.cull_timeout
    solara.server.settings.kernel.cull_timeout = "0.2s"
    try:
        yield
    finally:
        solara.server.settings.kernel.cull_timeout = cull_timeout_previous


@pytest.mark.skipif(on_windows, reason="This test is flaky on Windows")
async def test_kernel_lifecycle_reconnect_simple(short_cull_timeout):
    # a reconnect should be possible within the reconnect window
    websocket = Mock()
    context = kernel_context.initialize_virtual_kernel("session-id-1", "kernel-id-1", websocket)
    context.page_connect("page-id-1")
    cull_task1 = context.page_disconnect("page-id-1")
    await asyncio.sleep(0.01)
    context.page_connect("page-id-2")
    # the new connect should cancel the first cull task
    with pytest.raises(asyncio.CancelledError):
        await cull_task1
    assert not context.closed_event.is_set()
    await context.page_disconnect("page-id-2")
    assert context.closed_event.is_set()


@pytest.mark.skipif(on_windows, reason="This test is flaky on Windows")
async def test_kernel_lifecycle_double_disconnect(short_cull_timeout):
    # a reconnect should be possible within the reconnect window
    websocket = Mock()
    context = kernel_context.initialize_virtual_kernel("session-id-1", "kernel-id-1", websocket)
    context.page_connect("page-id-1")
    cull_task1 = context.page_disconnect("page-id-1")

    # now after 0.1 we disconnect the 2nd time
    await asyncio.sleep(0.1)
    context.page_connect("page-id-2")
    cull_task2 = context.page_disconnect("page-id-2")
    t_disconnect_page_2 = time.time()
    t0_disconnect_page_2 = time.time()

    # go over the reconnect window of cull_task1 (with a 0.05 extra to make sure it is really over)
    # await asyncio.sleep(0.1 + 0.05)
    # but the first disconnect should not have closed the kernel context yet
    with pytest.raises(asyncio.CancelledError):
        await cull_task1
    assert (time.time() - t_disconnect_page_2) < 0.001, "should be cancelled really quickly"

    assert not context.closed_event.is_set()
    await cull_task2
    assert context.closed_event.is_set()
    # the context should be closed AFTER 0.2 seconds, but it could take a bit longer
    assert 0.3 >= (time.time() - t0_disconnect_page_2) >= 0.2


@pytest.mark.skipif(on_windows, reason="This test is flaky on Windows")
@pytest.mark.parametrize("close_first", [True, False])
async def test_kernel_lifecycle_close_single(close_first, short_cull_timeout):
    # a reconnect should be possible within the reconnect window
    websocket = Mock()
    context = kernel_context.initialize_virtual_kernel("session-id-1", "kernel-id-1", websocket)
    context.page_connect("page-id-1")
    if close_first:
        context.page_close("page-id-1")
        assert context.closed_event.is_set()
        context.page_disconnect("page-id-1")
    else:
        context.page_disconnect("page-id-1")
        assert not context.closed_event.is_set()
        context.page_close("page-id-1")
        assert context.closed_event.is_set()


@pytest.mark.skipif(on_windows, reason="This test is flaky on Windows")
@pytest.mark.parametrize("close_first", [True, False])
async def test_kernel_lifecycle_close_while_disconnected(close_first, short_cull_timeout):
    # a reconnect should be possible within the reconnect window
    websocket = Mock()
    context = kernel_context.initialize_virtual_kernel(f"session-id-1-{close_first}", f"kernel-id-1-{close_first}", websocket)
    context.page_connect("page-id-1")
    cull_task_1 = context.page_disconnect("page-id-1")
    await asyncio.sleep(0.1)
    # after 0.1 we connect again, but close it directly
    context.page_connect("page-id-2")
    if close_first:
        cull_task_2 = context.page_close("page-id-2")
        await asyncio.sleep(0.01)
        context.page_disconnect("page-id-2")
    else:
        context.page_disconnect("page-id-2")
        await asyncio.sleep(0.01)
        cull_task_2 = context.page_close("page-id-2")
    assert cull_task_2 is not None
    assert not context.closed_event.is_set()
    await asyncio.sleep(0.15)
    # but even though we closed, the first page is still in the disconnected state
    with pytest.raises(asyncio.CancelledError):
        await cull_task_1
    assert not context.closed_event.is_set()
    await cull_task_2
    assert context.closed_event.is_set()



import logging
import threading
import time
from typing import Optional
import socket

logger = logging.getLogger("solara.server.threaded")


def get_free_port():
    sock = socket.socket()
    sock.bind(("", 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


class ServerBase(threading.Thread):
    name = "not set"

    def __init__(self, port: int, host: str = "localhost", **kwargs):
        self.port = get_free_port() if port == 0 else port
        self.host = host
        self.base_url = f"http://{self.host}:{self.port}"

        self.kwargs = kwargs
        self.started = threading.Event()
        self.stopped = threading.Event()
        self.error: Optional[BaseException] = None
        super().__init__(name="test-server-thread", daemon=True)

    def run(self):
        try:
            logger.info("Starting main loop")
            self.mainloop()
        except BaseException as e:  # noqa
            self.error = e
            self.started.set()
            logger.exception("Issue starting server")

    def serve_threaded(self):
        logger.debug("start thread")
        self.start()
        logger.debug("wait for thread to run")
        self.started.wait()
        if self.error:
            raise self.error

    def wait_until_serving(self, timeout: float = 10):
        start = time.time()
        while time.time() < start + timeout:
            if self.has_started():
                time.sleep(0.1)  # give some time to really start
                return
            time.sleep(0.05)
        raise RuntimeError(f"Server at {self.base_url} does not seem to be running")

    def serve(self):
        raise NotImplementedError

    def mainloop(self):
        logger.info("serving at http://%s:%d" % (self.host, self.port))
        try:
            self.serve()
        except:  # noqa: E722
            logger.exception("Oops, server stopped unexpectedly")
            raise
        finally:
            self.stopped.set()

    def stop_serving(self):
        logger.debug("stopping server")
        self.signal_stop()
        self.stopped.wait(10)
        if not self.stopped.is_set():
            logger.error("stopping server failed")
        else:
            logger.debug("stopped server")

    def signal_stop(self):
        pass

    def has_started(self) -> bool:
        return False



import ast
import inspect
import re
import typing as t
from enum import Enum
import sys
import warnings

DEFAULT_USE_FUNCTIONS = ("^use_.*$",)


class InvalidReactivityCause(Enum):
    USE_AFTER_RETURN = "early return"
    CONDITIONAL_USE = "conditional"
    LOOP_USE = "loop"
    NESTED_FUNCTION_USE = "nested function"
    VARIABLE_ASSIGNMENT = "assignment"
    EXCEPTION_USE = "exception"


noqa_code_to_cause = {
    "SH101": InvalidReactivityCause.USE_AFTER_RETURN,
    "SH102": InvalidReactivityCause.CONDITIONAL_USE,
    "SH103": InvalidReactivityCause.LOOP_USE,
    "SH104": InvalidReactivityCause.NESTED_FUNCTION_USE,
    "SH105": InvalidReactivityCause.VARIABLE_ASSIGNMENT,
    "SH106": InvalidReactivityCause.EXCEPTION_USE,
}
cause_to_noqa_code = {v: k for k, v in noqa_code_to_cause.items()}
noqa_pattern = re.compile(r".*# noqa(?::\s*([A-Z]{2,3}\d{2,3})(?:,\s*([A-Z]{2,3}\d{2,3}))*)?")


if sys.version_info < (3, 11):
    ScopeNodeType = t.Union[ast.For, ast.While, ast.If, ast.Try, ast.FunctionDef]
    TryNodes = (ast.Try,)
else:
    # except* nodes are only standardized in 3.11+
    ScopeNodeType = t.Union[ast.For, ast.While, ast.If, ast.Try, ast.TryStar, ast.FunctionDef]
    TryNodes = (ast.Try, ast.TryStar)


def line_to_noqa(line: str) -> t.Optional[t.Set[InvalidReactivityCause]]:
    """Return set of noqa code for which to not do quality assurance checks, or None when to do all qa."""
    if "#" in line:
        no_qa = set()
        match = noqa_pattern.match(line)
        if match is not None:
            # not sure why we get some None values
            groups = [k for k in match.groups() if k is not None]
            if groups:  # `noqa: SHXXX`` found
                for group in groups:
                    if group not in noqa_code_to_cause:
                        if group.startswith("SH"):
                            raise ValueError(f"Unknown noqa code {group}")
                    else:
                        no_qa.add(noqa_code_to_cause[group])
            else:  # we found `noqa`
                # skip all qa
                no_qa = set(noqa_code_to_cause.values())
            return no_qa
        else:
            return None  # no noqa found
    return None  # fast path, no comment found


def should_skip_qa(global_no_qa: t.Optional[t.Set[InvalidReactivityCause]], cause: InvalidReactivityCause, line: str) -> bool:
    line_qa = line_to_noqa(line)
    if global_no_qa is None and line_qa is None:
        # there is not a noqa on the function or line
        return False
    else:
        noqa = global_no_qa or set()
        if line_qa is not None:
            noqa |= line_qa
        return cause in noqa


class HookValidationError(Exception):
    def __init__(self, cause: InvalidReactivityCause, message: str):
        self.cause = cause
        super().__init__(message)


class HookValidator(ast.NodeVisitor):
    def __init__(self, component: t.Callable, use_functions=DEFAULT_USE_FUNCTIONS):
        self.use_functions = [re.compile(use_func) for use_func in use_functions]

        self.root_function_return: t.Optional[ast.Return] = None
        self.outer_scope: t.Optional[ScopeNodeType] = None

        self.filename = component.__code__.co_filename
        self.line_offset = component.__code__.co_firstlineno - 1
        self.component = component

        self.source = inspect.getsource(self.component)
        # lines before we dedent
        self.lines = self.source.split("\n")
        # dedent the source code to avoid indentation errors
        parsed_source = inspect.cleandoc(self.source)
        parsed = ast.parse(parsed_source)
        # Get nodes from inside the function body
        func_definition = t.cast(ast.FunctionDef, parsed.body[0])
        self.function_scope: ast.FunctionDef = func_definition
        self._root_function_scope = self.function_scope
        # None means, *DO* qa
        self.no_qa: t.Optional[t.Set[InvalidReactivityCause]] = None

    def run(self):
        import solara.settings

        if solara.settings.main.check_hooks == "off":
            return
        func_def_line = self.lines[self._root_function_scope.lineno - 1]
        self.no_qa = line_to_noqa(func_def_line)
        try:
            for node in self._root_function_scope.body:
                self.visit(node)
        except HookValidationError as e:
            if solara.settings.main.check_hooks == "error":
                raise e
            elif solara.settings.main.check_hooks == "warn":
                warnings.warn(str(e))

    def matches_use_function(self, name: str) -> bool:
        return any(use_func.match(name) for use_func in self.use_functions)

    def node_to_scope_cause(self, node: ScopeNodeType) -> InvalidReactivityCause:
        if isinstance(node, ast.If):
            return InvalidReactivityCause.CONDITIONAL_USE
        elif isinstance(node, (ast.For, ast.While)):
            return InvalidReactivityCause.LOOP_USE
        elif isinstance(node, ast.FunctionDef):
            return InvalidReactivityCause.NESTED_FUNCTION_USE
        elif isinstance(node, TryNodes):
            return InvalidReactivityCause.EXCEPTION_USE
        else:
            warnings.warn(f"Unexpected scope node type: {node}, line={node.lineno}")

    def visit_Call(self, node: ast.Call):
        """Records calls of use functions, i.e. solara.use_state(...)"""
        func = node.func
        if isinstance(func, ast.Call):
            # Nested function, it will appear in another node later
            return
        id_ = None
        if isinstance(func, ast.Name):
            id_ = func.id
        elif isinstance(func, ast.Attribute):
            id_ = func.attr
        if id_ is not None and self.matches_use_function(id_):
            self.error_on_early_return(node, id_)
            self.error_on_invalid_scope(node, id_)
        self.generic_visit(node)

    def visit_If(self, node: ast.If):
        self._visit_children_using_scope(node)

    def visit_For(self, node: ast.For):
        self._visit_children_using_scope(node)

    def visit_While(self, node: ast.While):
        self._visit_children_using_scope(node)

    def visit_Try(self, node: ast.Try) -> t.Any:
        self._visit_children_using_scope(node)

    if sys.version_info >= (3, 11):

        def visit_TryStar(self, node: ast.TryStar) -> t.Any:
            self._visit_children_using_scope(node)

    def visit_FunctionDef(self, node: ast.FunctionDef):
        old_function_scope = self.function_scope
        self.function_scope = node
        self._visit_children_using_scope(node)
        self.function_scope = old_function_scope

    def _visit_children_using_scope(self, node: ScopeNodeType):
        outer_scope = self.outer_scope
        self.outer_scope = node
        for child in node.body:
            self.visit(child)
        self.outer_scope = outer_scope

    def visit_Assign(self, node: ast.Assign):
        self.error_on_invalid_assign(node)
        self.generic_visit(node)

    def visit_Return(self, node: ast.Return):
        """
        Records the earliest return statement in the function
        """
        # Returns are valid in nested functions
        if self.function_scope != self._root_function_scope:
            return
        self.root_function_return = node
        self.generic_visit(node)

    def error_on_invalid_assign(self, node: ast.Assign):
        if isinstance(node.value, ast.Attribute) and self.matches_use_function(node.value.attr):
            line = node.lineno + self.line_offset
            cause = InvalidReactivityCause.VARIABLE_ASSIGNMENT
            message = f"Assigning the hook `{node.value.attr}` to a variable is not allowed since it complicates the tracking of valid hook use {_hint_supress(line, cause)}"

            raise HookValidationError(
                cause,
                f"{self.get_source_context(line)}: {message}",
            )

    def error_on_early_return(self, use_node: ast.Call, use_node_id: str):
        """
        Checks if the latest use of a reactive function occurs after the earliest return
        """
        if self.root_function_return and self.root_function_return.lineno <= use_node.lineno:
            offset_return = self.root_function_return.lineno + self.line_offset
            offset_use = use_node.lineno + self.line_offset
            cause = InvalidReactivityCause.USE_AFTER_RETURN
            line = self.lines[use_node.lineno - 1]
            if should_skip_qa(self.no_qa, cause, line):
                return

            raise HookValidationError(
                cause,
                f"""{self.get_source_context(offset_use)}: `{use_node_id}` found despite early return on line {offset_return}
{_hint_supress(line, cause)}""",
            )

    def error_on_invalid_scope(self, use_node: ast.Call, use_node_id: str):
        """
        Checks if the latest use of a reactive function occurs after the earliest return or in an invalid scope
        such as try-except blocks.
        """
        if self.outer_scope is None:
            return
        offset_use = use_node.lineno + self.line_offset
        cause = self.node_to_scope_cause(self.outer_scope)

        scope_line = self.outer_scope.lineno + self.line_offset
        line = self.lines[use_node.lineno - 1]
        if should_skip_qa(self.no_qa, cause, line):
            return
        raise HookValidationError(
            cause,
            f"""{self.get_source_context(offset_use)}: `{use_node_id}` found within a {cause.value} created on line {scope_line}
{_hint_supress(line, cause)}""",
        )

    def get_source_context(self, lineno):
        return f"{self.filename}:{lineno}: {self.component.__qualname__}"


def _hint_supress(line, cause):
    return f"""To suppress this check, replace the line with:
{line}  # noqa: {cause_to_noqa_code[cause]}

Make sure you understand the consequences of this, by reading about the rules of hooks at:
    https://solara.dev/documentation/advanced/understanding/rules-of-hooks
"""



"""
# display

"""

import solara
from solara.website.components import NoPage
from solara.website.utils import apidoc

title = "display"


Page = NoPage


__doc__ += apidoc(solara.display)  # type: ignore



import pytest

pytest.importorskip("vaex")
import vaex

import solara
import solara.components.pivot_table as pt
from solara.components.pivot_table import PivotTableWidget

from .common import repeat_while_false, repeat_while_true

df = vaex.datasets.titanic()


def test_df_data_pivot_data_vaex():
    dfg = pt.df_aggregate_vaex(df, ["sex"], {"count": solara.AggregationCount(type="count")})
    assert dfg["count"].tolist() == [466, 843]

    data = pt.df_aggregate_pivot_vaex(df, ["sex"], [], solara.AggregationCount(type="count"))
    assert data["values_x"] == ["466", "843"]
    assert data["values_y"] == []
    assert data["values"] == [[]]
    assert data["total"] == "1,309"

    data = pt.df_aggregate_pivot_vaex(df, [], ["sex"], solara.AggregationCount(type="count"))
    assert data["values_x"] == []
    assert data["values_y"] == ["466", "843"]
    assert data["values"] == [[]]
    assert data["total"] == "1,309"

    data = pt.df_aggregate_pivot_vaex(df, ["survived"], ["sex"], solara.AggregationCount(type="count"))
    assert data["values_x"] == ["809", "500"]
    assert data["values_y"] == ["466", "843"]
    assert data["values"] == [["127", "682"], ["339", "161"]]
    assert data["total"] == "1,309"


def test_df_data_pivot_table_view():
    data = pt.df_aggregate_pivot_vaex(df, ["survived"], ["sex"], solara.AggregationCount(type="count"))
    el = solara.PivotTableView(data=data)
    box, rc = solara.render(el, handle_error=False)
    assert rc._find(PivotTableWidget).widget.d["values"] == [["127", "682"], ["339", "161"]]


def test_df_data_pivot_table_df():
    el = solara.PivotTable(df, ["survived"], ["sex"])
    box, rc = solara.render(el, handle_error=False)
    repeat_while_false(lambda: rc._find(PivotTableWidget))
    assert rc.find(PivotTableWidget).widget.d["values"] == [["127", "682"], ["339", "161"]]

    el = solara.PivotTableCard(df, ["survived"], ["sex"])
    box, rc = solara.render(el, handle_error=False)
    repeat_while_false(lambda: rc._find(PivotTableWidget))
    assert rc.find(PivotTableWidget).widget.d["values"] == [["127", "682"], ["339", "161"]]


def test_pivot_table():
    filter = set_filter = None

    @solara.component
    def FilterDummy(df):
        nonlocal filter, set_filter
        filter, set_filter = solara.use_cross_filter(id(df), "test")
        return solara.Text("dummy")

    @solara.component
    def Test():
        solara.provide_cross_filter()
        with solara.VBox() as main:
            solara.PivotTableCard(df, x=["sex"], y=["survived"])
            FilterDummy(df)
        return main

    widget, rc = solara.render(Test(), handle_error=False)
    rc.find(PivotTableWidget).wait_for(timeout=10)
    pt = rc._find(PivotTableWidget).widget
    data = pt.d
    assert data["x"] == ["sex"]
    assert data["y"] == ["survived"]
    assert data["values_x"] == ["466", "843"]
    assert data["values_y"] == ["809", "500"]
    assert data["values"] == [["127", "339"], ["682", "161"]]
    assert data["total"] == f"{len(df):,}"
    assert set_filter is not None
    set_filter(str(df["pclass"] == 2))
    # wait for the data to change
    rc.find(PivotTableWidget).assert_wait(lambda w: w.d["values"] != [["127", "339"], ["682", "161"]], timeout=10)
    data = pt.d
    assert data["values_x"] == ["106", "171"]
    assert data["values_y"] == ["158", "119"]
    assert data["values"] == [["12", "94"], ["146", "25"]]
    assert data["total"] == f"{len(df[df.pclass == 2]):,}"
    set_filter(None)
    # wait for the filter to be applied (data should change)
    rc.find(PivotTableWidget).assert_wait(lambda w: w.d["values"] != [["12", "94"], ["146", "25"]], timeout=10)
    pt.selected = {"x": [0, 0]}  # sex, female
    repeat_while_true(lambda: filter is None)
    assert df[df[str(filter)]].sex.unique() == ["female"]

    pt.selected = {"y": [0, 0]}  # survived, False
    assert filter is not None
    assert df[df[str(filter)]].survived.unique() == [False]



import asyncio
import sys

try:
    import contextvars
except ModuleNotFoundError:
    contextvars = None  # type: ignore

import concurrent.futures
import contextlib
import dataclasses
import enum
import logging
import os
import pickle
import threading
import time
import typing
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, cast

import ipywidgets as widgets
import reacton
from ipywidgets import DOMWidget, Widget

import solara.server.settings
import solara.util

from . import kernel, websocket
from .. import lifecycle
from .kernel import Kernel, WebsocketStreamWrapper

WebSocket = Any
logger = logging.getLogger("solara.server.app")


class Local(threading.local):
    kernel_context_stack: Optional[List[Optional["VirtualKernelContext"]]] = None


local = Local()


class PageStatus(enum.Enum):
    CONNECTED = "connected"
    DISCONNECTED = "disconnected"
    CLOSED = "closed"


@dataclasses.dataclass
class VirtualKernelContext:
    id: str
    kernel: kernel.Kernel
    # we keep track of the session id to prevent kernel hijacking
    # to 'steal' a kernel, one would need to know the session id
    # *and* the kernel id
    session_id: str
    control_sockets: List[WebSocket] = dataclasses.field(default_factory=list)
    # this is the 'private' version of the normally global ipywidgets.Widgets.widget dict
    # see patch.py
    widgets: Dict[str, Widget] = dataclasses.field(default_factory=dict)
    # same, for ipyvue templates
    # see patch.py
    templates: Dict[str, Widget] = dataclasses.field(default_factory=dict)
    user_dicts: Dict[str, Dict] = dataclasses.field(default_factory=dict)
    # anything we need to attach to the context
    # e.g. for a react app the render context, so that we can store/restore the state
    app_object: Optional[Any] = None
    reload: Callable = lambda: None  # noqa: E731
    state: Any = None
    container: Optional[DOMWidget] = None
    # we track which pages are connected to implement kernel culling
    page_status: Dict[str, PageStatus] = dataclasses.field(default_factory=dict)
    # only used for testing
    _last_kernel_cull_task: "Optional[asyncio.Future[None]]" = None
    _last_kernel_cull_future: "Optional[concurrent.futures.Future[None]]" = None
    closed_event: threading.Event = dataclasses.field(default_factory=threading.Event)
    _on_close_callbacks: List[Callable[[], None]] = dataclasses.field(default_factory=list)
    lock: threading.RLock = dataclasses.field(default_factory=threading.RLock)
    event_loop: asyncio.AbstractEventLoop = dataclasses.field(default_factory=asyncio.get_event_loop)

    def __post_init__(self):
        with self:
            for f, *_ in lifecycle._on_kernel_start_callbacks:
                cleanup = f()
                if cleanup:
                    self.on_close(cleanup)

    def restart(self):
        # should we do this, or maybe close the context and create a new one?
        with self:
            for f in reversed(self._on_close_callbacks):
                f()
            self._on_close_callbacks.clear()
            self.__post_init__()

    def display(self, *args):
        print(args)  # noqa

    def on_close(self, f: Callable[[], None]):
        self._on_close_callbacks.append(f)

    def __enter__(self):
        if local.kernel_context_stack is None:
            local.kernel_context_stack = []
        key = get_current_thread_key()
        local.kernel_context_stack.append(current_context.get(key, None))
        current_context[key] = self

    def __exit__(self, *args):
        key = get_current_thread_key()
        assert local.kernel_context_stack is not None
        current_context[key] = local.kernel_context_stack.pop()

    def close(self):
        with self, self.lock:
            for key in self.page_status:
                self.page_status[key] = PageStatus.CLOSED
            if self._last_kernel_cull_task:
                self._last_kernel_cull_task.cancel()
            if self._last_kernel_cull_future:
                self._last_kernel_cull_future.cancel()
            if self.closed_event.is_set():
                logger.error("Tried to close a kernel context that is already closed: %s", self.id)
                return
            logger.info("Shut down virtual kernel: %s", self.id)
            for f in reversed(self._on_close_callbacks):
                f()
            if self.app_object is not None:
                if isinstance(self.app_object, reacton.core._RenderContext):
                    try:
                        self.app_object.close()
                    except Exception as e:
                        logger.exception("Could not close render context: %s", e)
                        # we want to continue, so we at least close all widgets
            widgets.Widget.close_all()
            # what if we reference each other
            # import gc
            # gc.collect()
            self.kernel.close()
            self.kernel = None  # type: ignore
            if self.id in contexts:
                del contexts[self.id]
            del current_context[get_current_thread_key()]
            # We saw in memleak_test that there are sometimes other entries in current_context
            # In which _DummyThread's reference this context, so we remove those references too
            # TODO: Think about what to do with those Threads
            _contexts = current_context.copy()
            for key, _ctx in _contexts.items():
                if _ctx is self:
                    del current_context[key]
            self.closed_event.set()

    def _state_reset(self):
        state_directory = Path(".") / "states"
        state_directory.mkdir(exist_ok=True)
        path = state_directory / f"{self.id}.pickle"
        path = path.absolute()
        try:
            path.unlink()
        except:  # noqa
            pass
        del contexts[self.id]
        key = get_current_thread_key()
        del current_context[key]

    def state_save(self, state_directory: os.PathLike):
        path = Path(state_directory) / f"{self.id}.pickle"
        render_context = self.app_object
        if render_context is not None:
            render_context = cast(reacton.core._RenderContext, render_context)
            state = render_context.state_get()
            with path.open("wb") as f:
                logger.debug("State: %r", state)
                pickle.dump(state, f)

    def page_connect(self, page_id: str):
        if self.closed_event.is_set():
            raise RuntimeError("Cannot connect a page to a closed kernel")
        logger.info("Connect page %s for kernel %s", page_id, self.id)
        with self.lock:
            if self.closed_event.is_set():
                raise RuntimeError("Cannot connect a page to a closed kernel")
            if page_id in self.page_status and self.page_status.get(page_id) == PageStatus.CLOSED:
                raise RuntimeError("Cannot connect a page that is already closed")
            self.page_status[page_id] = PageStatus.CONNECTED
            if self._last_kernel_cull_task:
                logger.info("Cancelling previous kernel cull task for virtual kernel %s", self.id)
                self._last_kernel_cull_task.cancel()

    def _bump_kernel_cull(self):
        async def kernel_cull():
            try:
                cull_timeout_sleep_seconds = solara.util.parse_timedelta(solara.server.settings.kernel.cull_timeout)
                logger.info("Scheduling kernel cull, will wait for max %s before shutting down the virtual kernel %s", cull_timeout_sleep_seconds, self.id)
                await asyncio.sleep(cull_timeout_sleep_seconds)
                logger.info("Timeout reached, checking if we should be shutting down virtual kernel %s", self.id)
                with self.lock:
                    has_connected_pages = PageStatus.CONNECTED in self.page_status.values()
                    if has_connected_pages:
                        logger.info("We have (re)connected pages, keeping the virtual kernel %s alive", self.id)
                    else:
                        logger.info("No connected pages, and timeout reached, shutting down virtual kernel %s", self.id)
                        self.close()
                if current_event_loop is not None and future is not None:
                    try:
                        current_event_loop.call_soon_threadsafe(future.set_result, None)
                    except RuntimeError:
                        pass  # event loop already closed, happens during testing
            except asyncio.CancelledError:
                if current_event_loop is not None and future is not None:
                    try:
                        if sys.version_info >= (3, 9):
                            current_event_loop.call_soon_threadsafe(future.cancel, "cancelled because a new cull task was scheduled")
                        else:
                            current_event_loop.call_soon_threadsafe(future.cancel)
                    except RuntimeError:
                        pass  # event loop already closed, happens during testing
                raise

        async def create_task():
            task = asyncio.create_task(kernel_cull())
            # create a reference to the task so we can cancel it later
            self._last_kernel_cull_task = task
            await task

        with self.lock:
            future: "Optional[asyncio.Future[None]]" = None
            current_event_loop: Optional[asyncio.AbstractEventLoop] = None
            try:
                future = asyncio.Future()
                current_event_loop = asyncio.get_event_loop()
            except RuntimeError:
                pass
            if self._last_kernel_cull_task:
                logger.info("Cancelling previous kernel cull tas for virtual kernel %s", self.id)
                self._last_kernel_cull_task.cancel()

            logger.info("Scheduling kernel cull for virtual kernel %s", self.id)

            async def create_task():
                task = asyncio.create_task(kernel_cull())
                # create a reference to the task so we can cancel it later
                self._last_kernel_cull_task = task
                try:
                    await task
                except RuntimeError:
                    pass  # event loop already closed, happens during testing

            self._last_kernel_cull_future = asyncio.run_coroutine_threadsafe(create_task(), keep_alive_event_loop)
            return future

    def page_disconnect(self, page_id: str) -> "Optional[asyncio.Future[None]]":
        """Signal that a page has disconnected, and schedule a kernel cull if needed.

        During the kernel reconnect window, we will keep the kernel alive, even if all pages have disconnected.

        Will return a future that is set when the kernel cull is done, when an event loop is available.
        The scheduled kernel cull can be cancelled when a new page connects, a new disconnect is scheduled,
        or a page if explicitly closed.
        """

        logger.info("Disconnect page %s for kernel %s", page_id, self.id)
        future: "asyncio.Future[None]" = asyncio.Future()
        with self.lock:
            if self.page_status[page_id] == PageStatus.CLOSED:
                # this happens when the close beackon call happens before the websocket disconnect
                logger.info("Page %s already closed for kernel %s", page_id, self.id)
                future.set_result(None)
                return future
            assert self.page_status[page_id] == PageStatus.CONNECTED, "cannot disconnect a page that is in state: %r" % self.page_status[page_id]
            self.page_status[page_id] = PageStatus.DISCONNECTED
            has_connected_pages = PageStatus.CONNECTED in self.page_status.values()
            if not has_connected_pages:
                # when we have no connected pages, we will schedule a kernel cull
                future = self._bump_kernel_cull()
            else:
                logger.info("Still have connected pages, do nothing for kernel %s", self.id)
                future.set_result(None)
            return future

    def page_close(self, page_id: str):
        """Signal that a page has closed, close the context if needed and schedule a kernel cull if needed.

        Closing the browser tab or a page navigation means an explicit close, which is
        different from a websocket/page disconnect, which we might want to recover from.

        """
        future: "Optional[asyncio.Future[None]]" = None

        try:
            future = asyncio.Future()
        except RuntimeError:
            pass
        else:
            future.set_result(None)

        logger.info("page status: %s", self.page_status)
        with self.lock:
            if self.closed_event.is_set():
                logger.info("Kernel %s was already closed when page %s attempted to close", self.id, page_id)
                return
            if self.page_status[page_id] == PageStatus.CLOSED:
                logger.info("Page %s already closed for kernel %s", page_id, self.id)
                return
            self.page_status[page_id] = PageStatus.CLOSED
            logger.info("Close page %s for kernel %s", page_id, self.id)
            has_connected_pages = PageStatus.CONNECTED in self.page_status.values()
            has_disconnected_pages = PageStatus.DISCONNECTED in self.page_status.values()
            # if we have disconnected pages, we may have cancelled the kernel cull task
            # if we still have connected pages, it will go to a disconnected state again
            # which will also trigger a new kernel cull
            if has_disconnected_pages:
                future = self._bump_kernel_cull()
            if not (has_connected_pages or has_disconnected_pages):
                logger.info("No connected or disconnected pages, shutting down virtual kernel %s", self.id)
                self.close()
            else:
                logger.info("Still have connected or disconnected pages, keeping virtual kernel %s alive", self.id)
        return future


try:
    # Normal Python
    keep_alive_event_loop = asyncio.new_event_loop()

    def _run():
        asyncio.set_event_loop(keep_alive_event_loop)
        try:
            keep_alive_event_loop.run_forever()
        except Exception:
            logger.exception("Error in keep alive event loop")
            raise

    threading.Thread(target=_run, daemon=True).start()
except RuntimeError:
    # Emscripten/pyodide/lite
    keep_alive_event_loop = asyncio.get_event_loop()

contexts: Dict[str, VirtualKernelContext] = {}
# maps from thread key to VirtualKernelContext, if VirtualKernelContext is None, it exists, but is not set as current
current_context: Dict[str, Optional[VirtualKernelContext]] = {}


def create_dummy_context():
    from . import kernel

    kernel_context = VirtualKernelContext(
        id="dummy",
        session_id="dummy",
        kernel=kernel.Kernel(),
    )
    return kernel_context


if contextvars is not None:
    if typing.TYPE_CHECKING:
        async_context_id = contextvars.ContextVar[str]("async_context_id")
    else:
        async_context_id = contextvars.ContextVar("async_context_id")
    async_context_id.set("default")
else:
    async_context_id = None


def get_current_thread_key() -> str:
    if not solara.server.settings.kernel.threaded:
        if async_context_id is not None:
            try:
                key = async_context_id.get()
            except LookupError:
                raise RuntimeError("no kernel context set")
        else:
            raise RuntimeError("No threading support, and no contextvars support (Python 3.6 is not supported for this)")
    else:
        thread = threading.current_thread()
        key = get_thread_key(thread)
    return key


def get_thread_key(thread: threading.Thread) -> str:
    if not solara.server.settings.kernel.threaded:
        if async_context_id is not None:
            return async_context_id.get()
    thread_key = thread._name + str(thread._ident)  # type: ignore
    return thread_key


def set_context_for_thread(context: VirtualKernelContext, thread: threading.Thread):
    key = get_thread_key(thread)
    current_context[key] = context


def clear_context_for_thread(thread: threading.Thread):
    key = get_thread_key(thread)
    current_context.pop(key, None)


def has_current_context() -> bool:
    thread_key = get_current_thread_key()
    return (thread_key in current_context) and (current_context[thread_key] is not None)


def get_current_context() -> VirtualKernelContext:
    thread_key = get_current_thread_key()
    if thread_key not in current_context:
        raise RuntimeError(
            f"Tried to get the current context for thread {thread_key}, but no known context found. This might be a bug in Solara. "
            f"(known contexts: {list(current_context.keys())}"
        )
    context = current_context[thread_key]
    if context is None:
        raise RuntimeError(
            f"Tried to get the current context for thread {thread_key!r}, although the context is know, it was not set for this thread. "
            + "This might be a bug in Solara."
        )
    return context


def set_current_context(context: Optional[VirtualKernelContext]):
    thread_key = get_current_thread_key()
    current_context[thread_key] = context


@contextlib.contextmanager
def without_context():
    context = None
    try:
        context = get_current_context()
    except RuntimeError:
        pass
    thread_key = get_current_thread_key()
    current_context[thread_key] = None
    try:
        yield
    finally:
        current_context[thread_key] = context


def initialize_virtual_kernel(session_id: str, kernel_id: str, websocket: websocket.WebsocketWrapper):
    from solara.server import app as appmodule

    if kernel_id in contexts:
        logger.info("reusing virtual kernel: %s", kernel_id)
        context = contexts[kernel_id]
        if context.session_id != session_id:
            logger.critical("Session id mismatch when reusing kernel (hack attempt?): %s != %s", context.session_id, session_id)
            websocket.send_text("Session id mismatch when reusing kernel (hack attempt?)")
            # to avoid very fast reconnects (we are in a thread anyway)
            time.sleep(0.5)
            raise ValueError("Session id mismatch")
        kernel = context.kernel
    else:
        kernel = Kernel()
        logger.info("new virtual kernel: %s", kernel_id)
        context = contexts[kernel_id] = VirtualKernelContext(id=kernel_id, session_id=session_id, kernel=kernel, control_sockets=[], widgets={}, templates={})

        with context:
            widgets.register_comm_target(kernel)
            appmodule.register_solara_comm_target(kernel)
    with context:
        assert has_current_context()
        assert kernel is Kernel.instance()
        kernel.shell_stream = WebsocketStreamWrapper(websocket, "shell")
        kernel.control_stream = WebsocketStreamWrapper(websocket, "control")
        kernel.session.websockets.add(websocket)
    return context



from typing import Dict, List, Union

import ipyvue as vue
import reacton.ipyvue as vuer
import solara


@solara.component
def Link(
    path_or_route: Union[str, solara.Route],
    children=[],
    nofollow=False,
    style: Union[str, Dict[str, str], None] = None,
    classes: List[str] = [],
):
    """Makes clicking on child elements navigate to a route.

    See also:

     * [Multipage](/documentation/advanced/howto/multipage).
     * [Understanding Routing](/documentation/advanced/understanding/routing).

    Most common usage is in combination with a button, e.g.:

    ```python
    with solara.Link("/fruit/banana"):
        solara.Button("Go to banana")
    ```


    ## Arguments

     * path_or_route: the path or route to navigate to. Paths should be absolute, e.g. '/fruit/banana'.
       If a route is given, [`resolve_path`](/documentation/api/routing/resolve_path)] will be used to resolve to the absolute path.
     * children: the children of the link. If a child is clicked, the link will be followed.
     * nofollow: If True, the link will not be followed by web crawlers (such as google).
     * style: CSS styles to apply to the HTML link element. Either a string or a dictionary.
     * classes: A list of CSS classes to apply to the link.

    """
    path = solara.resolve_path(path_or_route, level=0)
    attributes = {"href": path}
    if nofollow:
        attributes["rel"] = "nofollow"
    style_flat = solara.util._flatten_style(style)
    classes_flat = solara.util._combine_classes(classes)

    link = vue.Html.element(tag="a", children=children, attributes=attributes, style_=style_flat, class_=classes_flat)
    location = solara.use_context(solara.routing._location_context)

    def go(*ignore):
        location.pathname = path

    vuer.use_event(link, "click.prevent.stop", go)
    return link



import solara

clicks = solara.reactive(0)


@solara.component
def Page():
    color = "green"
    if clicks.value >= 5:
        color = "red"

    def increment():
        clicks.value += 1
        print("clicks", clicks)  # noqa

    solara.Button(label=f"Clicked: {clicks}", on_click=increment, color=color)



"""# use_task"""

import solara
import solara.autorouting
import solara.lab
from solara.website.components import NoPage
from solara.website.utils import apidoc

title = "use_task"
Page = NoPage
__doc__ += apidoc(solara.lab.use_task)  # type: ignore



"""
# Select components

Select comes in two flavours:

   * `Select` for a singular selection
   * `SelectMultiple` which allows for multiple selections


"""

import solara
from solara.website.components import NoPage
from solara.website.utils import apidoc

Page = NoPage


__doc__ += "# Select"
__doc__ += apidoc(solara.Select.f)  # type: ignore
__doc__ += "# SelectMultiple"
__doc__ += apidoc(solara.SelectMultiple.f)  # type: ignore



from types import FrameType, ModuleType
from typing import Callable, List, NamedTuple, Optional
from pathlib import Path
import inspect


class _on_kernel_callback_entry(NamedTuple):
    callback: Callable[[], Optional[Callable[[], None]]]
    callpoint: Optional[Path]
    module: Optional[ModuleType]
    cleanup: Callable[[], None]


_on_kernel_start_callbacks: List[_on_kernel_callback_entry] = []


def _find_root_module_frame() -> Optional[FrameType]:
    # basically the module where the call stack origined from
    current_frame = inspect.currentframe()
    root_module_frame = None

    while current_frame is not None:
        if current_frame.f_code.co_name == "<module>":
            root_module_frame = current_frame
            break
        current_frame = current_frame.f_back

    return root_module_frame


def on_kernel_start(f: Callable[[], Optional[Callable[[], None]]]) -> Callable[[], None]:
    root = _find_root_module_frame()
    path: Optional[Path] = None
    module: Optional[ModuleType] = None
    if root is not None:
        path_str = inspect.getsourcefile(root)
        module = inspect.getmodule(root)
        if path_str is not None:
            path = Path(path_str)

    def cleanup():
        return _on_kernel_start_callbacks.remove(kce)

    kce = _on_kernel_callback_entry(f, path, module, cleanup)
    _on_kernel_start_callbacks.append(kce)
    return cleanup



from typing import Callable, Dict, List, Optional, TypeVar, Union, cast, overload

import ipyvuetify as v
import reacton.core

import solara
from solara.alias import rv
from solara.util import _combine_classes

T = TypeVar("T")


@overload
def Select(
    label: str,
    values: List[T],
    value: None = ...,
    on_value: Optional[Callable[[Optional[T]], None]] = ...,
    dense: bool = ...,
    disabled: bool = ...,
    classes: List[str] = [],
    style: Union[str, Dict[str, str], None] = None,
) -> reacton.core.ValueElement[v.Select, T]: ...


@overload
def Select(
    label: str,
    values: List[T],
    value: T = ...,
    on_value: Optional[Callable[[T], None]] = ...,
    dense: bool = ...,
    disabled: bool = ...,
    classes: List[str] = [],
    style: Union[str, Dict[str, str], None] = None,
) -> reacton.core.ValueElement[v.Select, T]: ...


@overload
def Select(
    label: str,
    values: List[T],
    value: solara.Reactive[Optional[T]] = ...,
    on_value: Optional[Callable[[Optional[T]], None]] = ...,
    dense: bool = ...,
    disabled: bool = ...,
    classes: List[str] = [],
    style: Union[str, Dict[str, str], None] = None,
) -> reacton.core.ValueElement[v.Select, T]: ...


@overload
def Select(
    label: str,
    values: List[T],
    value: solara.Reactive[T] = ...,
    on_value: Optional[Callable[[T], None]] = None,
    dense: bool = ...,
    disabled: bool = ...,
    classes: List[str] = [],
    style: Union[str, Dict[str, str], None] = None,
) -> reacton.core.ValueElement[v.Select, T]: ...


@solara.value_component(None)
def Select(
    label: str,
    values: List[T],
    value: Union[None, T, solara.Reactive[T], solara.Reactive[Optional[T]]] = None,
    on_value: Union[None, Callable[[T], None], Callable[[Optional[T]], None]] = None,
    dense: bool = False,
    disabled: bool = False,
    classes: List[str] = [],
    style: Union[str, Dict[str, str], None] = None,
) -> reacton.core.ValueElement[v.Select, T]:
    """Select a single value from a list of values.

    ### Basic example:

    ```solara
    import solara

    foods = ["Kiwi", "Banana", "Apple"]
    food = solara.reactive("Banana")


    @solara.component
    def Page():
        solara.Select(label="Food", value=food, values=foods)
        solara.Markdown(f"**Selected**: {food.value}")
    ```

    ## Arguments

     * `label`: Label to display next to the select.
     * `value`: The currently selected value.
     * `values`: List of values to select from.
     * `on_value`: Callback to call when the value changes.
     * `dense`: Whether to use a denser style.
     * `disabled`: Whether the select widget allow user interaction
     * `classes`: List of CSS classes to apply to the select.
     * `style`: CSS style to apply to the select.

    """
    # next line is very hard to get right with typing
    # might need an overload on use_reactive, when value is None
    reactive_value = solara.use_reactive(value, on_value)  # type: ignore
    del value, on_value
    style_flat = solara.util._flatten_style(style)
    class_ = _combine_classes(classes)
    return cast(
        reacton.core.ValueElement[v.Select, T],
        rv.Select(
            v_model=reactive_value.value,
            on_v_model=reactive_value.set,
            items=values,
            label=label,
            dense=dense,
            disabled=disabled,
            class_=class_,
            style_=style_flat,
        ),
    )


@solara.value_component(None)
def SelectMultiple(
    label: str,
    values: List[T],
    all_values: List[T],
    on_value: Callable[[List[T]], None] = None,
    dense: bool = False,
    disabled: bool = False,
    classes: List[str] = [],
    style: Union[str, Dict[str, str], None] = None,
) -> reacton.core.ValueElement[v.Select, List[T]]:
    """Select multiple values from a list of values.

    ### Basic example:

    ```solara
    import solara

    all_languages = "Python C++ Java JavaScript TypeScript BASIC".split()
    languages = solara.reactive([all_languages[0]])


    @solara.component
    def Page():
        solara.SelectMultiple("Languages", languages, all_languages)
        solara.Markdown(f"**Selected**: {languages.value}")
    ```

    ## Arguments

     * `label`: Label to display next to the select.
     * `values`: List of currently selected values.
     * `all_values`: List of all values to select from.
     * `on_value`: Callback to call when the value changes.
     * `dense`: Whether to use a denser style.
     * `disabled`: Whether the select widget allow user interaction
     * `classes`: List of CSS classes to apply to the select.
     * `style`: CSS style to apply to the select.
    """
    reactive_values = solara.use_reactive(values, on_value)
    del values, on_value
    style_flat = solara.util._flatten_style(style)
    class_ = _combine_classes(classes)
    return cast(
        reacton.core.ValueElement[v.Select, List[T]],
        rv.Select(
            v_model=reactive_values.value,
            on_v_model=reactive_values.set,
            items=all_values,
            label=label,
            multiple=True,
            dense=dense,
            disabled=disabled,
            class_=class_,
            style_=style_flat,
        ),
    )



import json
import unittest.mock
from urllib.parse import unquote

import requests

import solara.server.telemetry


def test_telemetry_basic(mocker):
    post: unittest.mock.MagicMock = mocker.spy(requests, "post")
    solara.server.telemetry.track("test_event", {"test_prop": "test_value"})
    post.assert_called_once()
    data = json.loads(unquote(post.call_args[1]["data"])[5:])[0]  # type: ignore
    assert data["event"] == "test_event"


def test_telemetry_server_start_stopc(mocker):
    post: unittest.mock.MagicMock = mocker.spy(requests, "post")
    solara.server.telemetry.server_start()
    solara.server.telemetry.server_stop()
    post.assert_called()
    data = json.loads(unquote(post.call_args[1]["data"])[5:])[0]  # type: ignore
    assert data["event"] == "Solara server stop"
    assert data["properties"]["duration_seconds"] > 0



import solara
from solara.alias import rv, rvue


@solara.component
def LinkTab(path, label):
    # both adding href to tab or adding Link makes vuetify buggy
    with rv.Tab() as tab:
        # with solara.Link(path):
        solara.Text(label)
    router = solara.use_router()

    def go(*ignore):
        router.push(path)

    rvue.use_event(tab, "click.prevent.stop", go)
    return tab


@solara.component
def TabNavigation(children=[], vertical=False, **kwargs):
    children = children or []
    route_current, all_routes = solara.use_route()

    tab_index = all_routes.index(route_current) if route_current is not None else 0

    with rv.Tabs(v_model=tab_index, vertical=vertical, **kwargs) as main:
        for i, route in enumerate(all_routes):
            path = solara.resolve_path(route)
            LinkTab(path, route.label or "No title")
        if route_current is None:
            return solara.Error("Page does not exist")

        with rv.TabsItems(v_model=tab_index, children=children):
            pass

    return main



import json
from datetime import datetime
from unittest.mock import Mock

from solara.server.kernel import SessionWebsocket
import numpy as np


def test_session_datetime():
    # some libraries, such as plotly may put datetime objects in the json content
    class Dummy:
        pass

    websocket = Mock()
    stream = Dummy()
    stream.channel = "iopub"  # type: ignore
    session = SessionWebsocket()
    session.websockets.add(websocket)
    session.send(stream, {"msg_type": "test", "content": {"data": "test"}, "somedate": datetime.now()})  # type: ignore
    websocket.send.assert_called_once()


def test_numpy_scalar():
    class Dummy:
        pass

    websocket = Mock()
    stream = Dummy()
    stream.channel = "iopub"  # type: ignore
    session = SessionWebsocket()
    session.websockets.add(websocket)
    v = np.int64(42)
    session.send(stream, {"msg_type": "test", "content": {"a_numpy_scalar": v}})  # type: ignore
    websocket.send.assert_called_once()
    json_string = websocket.send.call_args[0][0]
    json_data = json.loads(json_string)
    assert json_data["content"]["a_numpy_scalar"] == 42



from PyInstaller.utils.hooks import collect_data_files, collect_submodules

hiddenimports = collect_submodules("solara")
datas = collect_data_files("solara")  # codespell:ignore datas
datas += collect_data_files("solara-ui")  # codespell:ignore datas

# this makes sure that inspect.getfile works, which is used for the
# vue component decorator
module_collection_mode = "pyz+py"



from pathlib import Path

from solara.website.components.markdown import MarkdownWithMetadata


title = "About Us"
HERE = Path(__file__)

Page = MarkdownWithMetadata(Path(HERE.parent / "about.md").read_text())



import solara


@solara.component
def Page():
    with solara.Link("/showcase"):
        solara.Text("« Back to Showcases")
    with solara.ColumnsResponsive(12, medium=6):
        solara.Markdown(
            """
        # Domino code assist

        [Domino Code Assist](https://www.dominodatalab.com/product/code-assist) (DCA) is a tool developed for [Domino Data Lab](https://www.dominodatalab.com/)
        which provides a simple, intuitive point-and-click interface for generating Python or R code.
        DCA uses many of the solara components to render the UI in the classical notebook and lab. Solara server is used to serve data apps.

        Domino code assist is deeply integrated in the Jupyter notebook.
        """
        )
        solara.Image("https://dxhl76zpt6fap.cloudfront.net/public/showcase/lca/load-data-lab.png", width="100%", classes=["pt-12"])

    with solara.ColumnsResponsive(12, medium=6):
        solara.Image("https://dxhl76zpt6fap.cloudfront.net/public/showcase/lca/viz.png", width="100%", classes=["pt-12"])
        solara.Markdown(
            """
        # Sophisticated UIs

        Using Solara, we can build sophisticated UIs that are easy to use, intuitive
        and have a modern look and feel.

        """
        )

    with solara.ColumnsResponsive(12, medium=6):
        solara.Markdown(
            """
        # Making apps.

        Using Solara we could build an "What You See Is What You Get" app designer.
        All solara components look and feel the same in the notebook and in the deployed app.

        """
        )
        solara.Image("https://dxhl76zpt6fap.cloudfront.net/public/showcase/lca/app-design.png", width="100%", classes=["pt-12"])

    with solara.ColumnsResponsive(12, medium=6):
        solara.Image("https://dxhl76zpt6fap.cloudfront.net/public/showcase/lca/app-deployed.png", width="100%", classes=["pt-12"])
        solara.Markdown(
            """

            # App deploying

            To deploy apps we use [solara server](/documentation/advanced/understanding/solara-server) which can
            render the same UI as in the notebook, but now sharable with non-technical users.

            Many users have deployed apps at Domino using Solara with
            a single click.

            """
        )



import uuid
from typing import Callable, Dict, List, Optional, Union

from typing_extensions import Literal

import solara
from solara.components.input import use_change


@solara.component
def ChatBox(
    children: List[solara.Element] = [],
    style: Optional[Union[str, Dict[str, str]]] = None,
    classes: List[str] = [],
):
    """
    The ChatBox component is a container for ChatMessage components.
    Its primary use is to ensure the proper ordering of messages,
    using `flex-direction: column-reverse` together with `reversed(messages)`.

    # Arguments

    * `children`: A list of child components.
    * `style`: CSS styles to apply to the component. Either a string or a dictionary.
    * `classes`: A list of CSS classes to apply to the component.
    """
    style_flat = solara.util._flatten_style(style)
    if "flex-grow" not in style_flat:
        style_flat += " flex-grow: 1;"
    if "flex-direction" not in style_flat:
        style_flat += " flex-direction: column-reverse;"
    if "overflow-y" not in style_flat:
        style_flat += " overflow-y: auto;"

    classes = [*classes, "chat-box"]
    with solara.Column(
        style=style_flat,
        classes=classes,
    ):
        for child in list(reversed(children)):
            solara.display(child)


@solara.component
def ChatInput(
    send_callback: Optional[Callable[[str], None]] = None,
    disabled: bool = False,
    disabled_input: bool = False,
    disabled_send: bool = False,
    style: Optional[Union[str, Dict[str, str]]] = None,
    autofocus: bool = False,
    input_text_style: Optional[Union[str, Dict[str, str]]] = None,
    classes: List[str] = [],
    input_text_classes: List[str] = [],
):
    """
    The ChatInput component renders a text input together with a send button.

    # Arguments

    * `send_callback`: A callback function for when the user presses enter or clicks the send button taking the message as an argument.
    * `disabled`: disable both input and send.
    * `disabled_input`: Whether the input should be disabled. Useful for disabling messages while a chatbot is replying.
    * `disabled_send`: Whether the send button should be disabled. Useful for disabling sending further messages while a chatbot is replying.
    * `style`: CSS styles to apply to the `solara.Row` containing the input field and submit button. Either a string or a dictionary.
    * `autofocus`: Determines if a component is to be autofocused or not (Default is False). Autofocus will occur during page load and only one component per page can have autofocus active.
    * `input_text_style`: CSS styles to apply to the `InputText` part of the component. Either a string or a dictionary.
    * `classes`: A list of CSS classes to apply to the component. Also applied to the container.
    * `input_text_classes`: A list of CSS classes to apply to the `InputText` part of the component.
    """
    message, set_message = solara.use_state("")  # type: ignore
    style_flat = solara.util._flatten_style(style)
    input_text_style_flat = solara.util._flatten_style(input_text_style)

    if "align-items" not in style_flat:
        style_flat += " align-items: center;"

    with solara.Row(style=style_flat, classes=classes):

        def send(*ignore_args):
            if message != "" and send_callback is not None:
                send_callback(message)
                set_message("")

        message_input = solara.v.TextField(
            label="Type a message...",
            v_model=message,
            on_v_model=set_message,
            rounded=True,
            filled=True,
            hide_details=True,
            autofocus=autofocus,
            style_="flex-grow: 1;" + input_text_style_flat,
            disabled=disabled or disabled_input,
            class_=" ".join(input_text_classes),
        )

        use_change(message_input, send, update_events=["keyup.enter"])

        button = solara.v.Btn(color="primary", icon=True, children=[solara.v.Icon(children=["mdi-send"])], disabled=message == "" or disabled or disabled_send)

        use_change(button, send, update_events=["click"])


@solara.component
def ChatMessage(
    children: Union[List[solara.Element], str] = [],
    user: bool = False,
    avatar: Union[solara.Element, str, Literal[False], None] = None,
    name: Optional[str] = None,
    color: Optional[str] = "rgba(0,0,0,.06)",
    avatar_background_color: Optional[str] = None,
    border_radius: Optional[str] = None,
    notch: bool = False,
    style: Optional[Union[str, Dict[str, str]]] = None,
    classes: List[str] = [],
):
    """
    The ChatMessage component renders a message. Messages with `user=True` are rendered on the right side of the screen,
    all others on the left.

    # Arguments

    * `children`: A list of child components.
    * `user`: Whether the message is from the current user or not.
    * `avatar`: An avatar to display next to the message. Can be a string representation of a URL or Material design icon name,
        a solara Element, False to disable avatars altogether, or None to display initials based on `name`.
    * `name`: The name of the user who sent the message.
    * `color`: The background color of the message. Defaults to `rgba(0,0,0,.06)`. Can be any valid CSS color.
    * `avatar_background_color`: The background color of the avatar. Defaults to `color` if left as `None`.
    * `border_radius`: Sets the roundness of the corners of the message. Defaults to `None`,
        which applies the default border radius of a `solara.Column`, i.e. `4px`.
    * `notch`: Whether to display a speech bubble style notch on the side of the message.
    * `style`: CSS styles to apply to the component. Either a string or a dictionary. Applied to the container of the message.
    * `classes`: A list of CSS classes to apply to the component. Applied to the same container.
    """
    style_flat = solara.util._flatten_style(style)

    if "border-radius" not in style_flat:
        style_flat += f" border-radius: {border_radius if border_radius is not None else ''};"
    if f"border-top-{'right' if user else 'left'}-radius" not in style_flat:
        style_flat += f" border-top-{'right' if user else 'left'}-radius: 0;"
    if "padding" not in style_flat:
        style_flat += " padding: .5em 1.5em;"

    msg_uuid = solara.use_memo(lambda: str(uuid.uuid4()), dependencies=[])
    with solara.Row(
        justify="end" if user else "start",
        style={"flex-direction": "row-reverse" if user else "row", "padding": "5px"},
    ):
        if avatar is not False:
            with solara.v.Avatar(color=avatar_background_color if avatar_background_color is not None else color):
                if avatar is None and name is not None:
                    initials = "".join([word[:1] for word in name.split(" ")])
                    solara.HTML(tag="span", unsafe_innerHTML=initials, classes=["headline"])
                elif isinstance(avatar, solara.Element):
                    solara.display(avatar)
                elif isinstance(avatar, str) and avatar.startswith("mdi-"):
                    solara.v.Icon(children=[avatar])
                else:
                    solara.HTML(tag="img", attributes={"src": avatar, "width": "100%"})
        classes_new = classes + ["chat-message-" + msg_uuid, "right" if user else "left"]
        with solara.Column(
            classes=classes_new,
            gap=0,
            style=style_flat,
        ):
            if name is not None:
                solara.Text(name, style="font-weight: bold;", classes=["message-name", "right" if user else "left"])
            for child in children:
                if isinstance(child, solara.Element):
                    solara.display(child)
                else:
                    solara.Markdown(child)
        # we use the uuid to generate 'scoped' CSS, i.e. css that only applies to the component instance.
        extra_styles = (
            f""".chat-message-{msg_uuid}:before{{
                content: '';
                position: absolute;
                width: 0;
                height: 0;
                border: 6px solid;
                top: 0;
            }}
            .chat-message-{msg_uuid}.left:before{{
                left: -12px;
                border-color: var(--color) var(--color) transparent transparent;
            }}
            .chat-message-{msg_uuid}.right:before{{
                right: -12px;
                border-color: var(--color) transparent transparent var(--color);
            }}"""
            if notch
            else ""
        )
        solara.Style(
            f"""
            .chat-message-{msg_uuid}{{
                --color: {color};
                max-width: 75%;
                position: relative;
            }}
            .chat-message-{msg_uuid}.left{{
                    border-top-left-radius: 0;
                    background-color:var(--color);
                    {"margin-left: 10px !important;" if notch else ""}
            }}
            .chat-message-{msg_uuid}.right{{
                    border-top-right-radius: 0;
                    background-color:var(--color);
                    {"margin-right: 10px !important;" if notch else ""}
            }}
            {extra_styles}
            """
        )



"""
# bqplot

[bqplot](https://github.com/bqplot/bqplot) is a 2-D visualization system for Jupyter, based on the constructs of the Grammar of Graphics.
"""

import numpy as np
import reacton.bqplot as bqplot

import solara

x0 = np.linspace(0, 2, 100)

exponent = solara.reactive(1.0)
log_scale = solara.reactive(False)


@solara.component
def Page(x=x0, ymax=5):
    y = x**exponent.value
    color = "red"
    display_legend = True
    label = "bqplot graph"

    solara.SliderFloat(value=exponent, min=0.1, max=3, label="Exponent")
    solara.Checkbox(value=log_scale, label="Log scale")

    x_scale = bqplot.LinearScale()
    if log_scale.value:
        y_scale = bqplot.LogScale(min=0.1, max=ymax)
    else:
        y_scale = bqplot.LinearScale(min=0, max=ymax)

    lines = bqplot.Lines(x=x, y=y, scales={"x": x_scale, "y": y_scale}, stroke_width=3, colors=[color], display_legend=display_legend, labels=[label])
    x_axis = bqplot.Axis(scale=x_scale)
    y_axis = bqplot.Axis(scale=y_scale, orientation="vertical")
    bqplot.Figure(axes=[x_axis, y_axis], marks=[lines], scale_x=x_scale, scale_y=y_scale, layout={"max_width": "100%", "width": "100%"})

    # return main



import solara

title = "use_exception"
set_fail = None
clear = None


@solara.component
def UnstableComponent(number: int):
    if number == 3:
        raise Exception("I do not like 3")
    return solara.Text(f"You picked {number}")


@solara.component
def Page():
    value, set_value = solara.use_state(1)
    value_previous = solara.use_previous(value)
    exception, clear_exception = solara.use_exception()
    # print(exception)
    if exception:

        def reset():
            set_value(value_previous)
            clear_exception()

        solara.Text("Exception: " + str(exception))
        solara.Button(label="Go to previous state", on_click=reset)
    else:
        solara.IntSlider(value=value, min=0, max=10, on_value=set_value, label="Pick a number, except 3")
        UnstableComponent(value)



"""
# FileDrop components

FileDrop comes in two flavours:

   * `FileDrop` for a single file upload
   * `FileDropMultiple` which allows for multiple file upload


"""

import textwrap
from typing import List, cast

import solara
from solara.components.file_drop import FileInfo
from solara.website.utils import apidoc


@solara.component
def FileDropMultipleDemo():
    content, set_content = solara.use_state(cast(List[bytes], []))
    filename, set_filename = solara.use_state(cast(List[str], []))
    size, set_size = solara.use_state(cast(List[int], []))

    def on_file(files: List[FileInfo]):
        set_filename([f["name"] for f in files])
        set_size([f["size"] for f in files])
        set_content([f["file_obj"].read(100) for f in files])

    solara.FileDropMultiple(
        label="Drag and drop files(s) here to read the first 100 bytes.",
        on_file=on_file,
        lazy=True,  # We will only read the first 100 bytes
    )
    if content:
        solara.Info(f"Number of uploaded files: {len(filename)}")
        for f, s, c in zip(filename, size, content):
            solara.Info(f"File {f} has total length: {s}\n, first 100 bytes:")
            solara.Preformatted("\n".join(textwrap.wrap(repr(c))))


@solara.component
def FileDropDemo():
    content, set_content = solara.use_state(b"")
    filename, set_filename = solara.use_state("")
    size, set_size = solara.use_state(0)

    def on_file(f: FileInfo):
        set_filename(f["name"])
        set_size(f["size"])
        set_content(f["file_obj"].read(100))

    solara.FileDrop(
        label="Drag and drop a file here to read the first 100 bytes.",
        on_file=on_file,
        lazy=True,  # We will only read the first 100 bytes
    )
    if content:
        solara.Info(f"File {filename} has total length: {size}\n, first 100 bytes:")
        solara.Preformatted("\n".join(textwrap.wrap(repr(content))))


@solara.component
def Page():
    with solara.Row():
        with solara.Card(title="FileDrop"):
            FileDropDemo()
        with solara.Card(title="FileDropMultiple"):
            FileDropMultipleDemo()


__doc__ += "# FileDrop"
__doc__ += apidoc(solara.FileDrop.f)  # type: ignore
__doc__ += "# FileDropMultiple"
__doc__ += apidoc(solara.FileDropMultiple.f)  # type: ignore



from typing import Any, Callable

import solara
import solara.widgets


@solara.component
def FigureAltair(
    chart,
    on_click: Callable[[Any], None] = None,
    on_hover: Callable[[Any], None] = None,
):
    """Renders an Altair chart using VegaLite.

    See also [our altair example](/documentation/examples/libraries/altair).

    ## Arguments

    - chart: Altair chart
    - on_click: Callback function for click events.
    - on_hover: Callback function for hover events.

    """
    import altair as alt

    with alt.renderers.enable("mimetype"):
        bundle = chart._repr_mimebundle_()[0]
        key4 = "application/vnd.vegalite.v4+json"
        key5 = "application/vnd.vegalite.v5+json"
        if key4 not in bundle and key5 not in bundle:
            raise KeyError(f"{key4} and {key5} not in mimebundle:\n\n{bundle}")
        spec = bundle.get(key5, bundle.get(key4))
        return solara.widgets.VegaLite.element(
            spec=spec, on_click=on_click, listen_to_click=on_click is not None, on_hover=on_hover, listen_to_hover=on_hover is not None
        )


# alias for backward compatibility
AltairChart = FigureAltair



import solara

github_url = solara.util.github_url(__file__)

# list of nice soft colors
colors = [
    "#e6194b",
    "#3cb44b",
    "#ffe119",
    "#4363d8",
    "#f58231",
    "#911eb4",
]


@solara.component
def Page1():
    with solara.Sidebar():
        solara.Button(label="View source", icon_name="mdi-github-circle", attributes={"href": github_url, "target": "_blank"}, text=True, outlined=True)
        solara.Markdown("The sidebar will get scrollbars automatically, independently of the main content.")
        for i in range(10):
            with solara.Card(f"Card {i}"):
                solara.Info(f"Text {i}", color=colors[i % len(colors)])

    solara.Markdown("The main content will get scrollbars automatically, independently of the sidebar.")
    for i in range(10):
        with solara.Card(f"Card {i}"):
            solara.Info(f"Text {i}", color=colors[i % len(colors)])


@solara.component
def Page2():
    # it is important we do not interrupt the height 100% chain
    limit_content_height = solara.use_reactive(True)
    # warning: if we add "margin": "10px" to the style, we will trigger
    # scrollbars in the parent div, if you really need to add margins,
    # you should correct for that in the height using "calc(100% - 20px)"
    with solara.Column(style={"height": "100%"}):
        with solara.Sidebar():
            solara.Button(label="View source", icon_name="mdi-github-circle", attributes={"href": github_url, "target": "_blank"}, text=True, outlined=True)
            solara.Markdown("Main content columns will scroll together, or independently if we limit height to 100%")
            solara.Checkbox(label="Limit content height", value=limit_content_height)

        # if we do not limit height to 100%, solara's AppLayout will add a scroll bar
        # to the main content, which will be independent of the sidebar's scroll bar
        # but both columns will scroll together
        with solara.Columns([2, 4], style={"height": "100%"} if limit_content_height.value else {}):
            with solara.Card("I have my own scroll bar"):
                solara.Markdown("")
                for i in range(10):
                    with solara.Card(f"Card {i}"):
                        solara.Info(f"Text {i}", color=colors[i % len(colors)])
            with solara.Card("I also have my own scroll bar"):
                solara.Markdown("")
                for i in range(20):
                    with solara.Card(f"Card {i}"):
                        solara.Info(f"Text {i}", color=colors[i % len(colors)])


routes = [
    solara.Route(path="/", component=Page1, label="Scrolling"),
    solara.Route(path="custom", component=Page2, label="Custom scrolling"),
]



import playwright
import playwright.sync_api


def test_landing(page_session: playwright.sync_api.Page, solara_server, solara_app):
    # with screenshot_on_error(page, 'tmp/test_docs_basics.png'):
    with solara_app("solara.website.pages"):
        page_session.goto(solara_server.base_url)
        page_session.get_by_role(role="heading", name="Build high-quality web applications in pure Python").wait_for()
        page_session.get_by_role("link", name="Documentation").first.click()
        page_session.locator("text=How to use our documentation").first.wait_for()
        page_session.go_back()
        page_session.get_by_role(role="heading", name="Build high-quality web applications in pure Python").wait_for()


def test_docs_basics(page_session: playwright.sync_api.Page, solara_server, solara_app):
    # with screenshot_on_error(page, 'tmp/test_docs_basics.png'):
    with solara_app("solara.website.pages"):
        page_session.goto(solara_server.base_url + "/documentation/api/routing/use_route/fruit/banana")
        page_session.locator("text=You chose banana").wait_for()
        page_session.locator('button:has-text("kiwi")').click()
        page_session.locator("text=You chose kiwi").wait_for()
        page_session.locator('button:has-text("apple")').click()
        page_session.locator("text=You chose apple").wait_for()
        # back to kiwi
        page_session.go_back()
        page_session.locator("text=You chose kiwi").wait_for()
        # back to banana
        page_session.go_back()
        page_session.locator("text=You chose banana").wait_for()

        # forward to kiwi
        page_session.go_forward()
        page_session.locator("text=You chose kiwi").wait_for()

        # go to wrong fruit
        page_session.locator('button:has-text("wrong fruit")').click()

        # and follow the fallback link
        page_session.locator("text=Fruit not found, go to banana").click()
        page_session.locator("text=You chose banana").wait_for()

        # another wrong link
        page_session.locator('button:has-text("wrong url")').click()
        page_session.locator("text=Page does not exist").wait_for()



import ipywidgets as widgets
import playwright.sync_api
from IPython.display import display
from typing import Callable
import time


def assert_equals_poll(getter: Callable, expected, timeout=2, iteration_delay=0.01):
    start = time.time()
    while time.time() - start < timeout:
        if getter() == expected:
            return
        time.sleep(iteration_delay)
    assert getter() == expected
    return False


def test_event_with_polling(solara_test, page_session: playwright.sync_api.Page):
    button = widgets.Button(description="Append data")
    # some data that will change due to a button click
    click_data = []

    def on_click(button):
        # change the data when the button is clicked
        # this will be called from the thread the websocket is in
        # so we can block/poll from the main thread (that pytest is running in)
        click_data.append(42)

    button.on_click(on_click)
    display(button)
    button_sel = page_session.locator("text=Append data")
    button_sel.click()

    # we block/poll until the condition is met.
    assert_equals_poll(lambda: click_data, [42])



import contextlib
import sys

import pytest
from dotenv import load_dotenv

import solara.server.reload
import solara.server.settings

load_dotenv()  # take environment variables from .env.  should be in os.environ for the whole test

solara.server.settings.telemetry.mixpanel_token = "adbf863d17cba80db608788e7fce9843"
solara.server.settings.main.mode = "development"
solara.server.reload.reloader.watcher = solara.server.reload.WatcherType([], solara.server.reload.reloader._on_change)


@pytest.fixture
def extra_include_path():
    @contextlib.contextmanager
    def extra_include_path(path):
        sys.path.insert(0, str(path))
        try:
            yield
        finally:
            sys.path[:] = sys.path[1:]

    return extra_include_path



import os
from typing import Dict, List, cast

import ipyvuetify as v
import ipywidgets
import traitlets

__all__ = [
    "VegaLite",
    "Navigator",
    "GridLayout",
    "HTML",
    "watch",
]


class VegaLite(v.VuetifyTemplate):
    template_file = os.path.realpath(os.path.join(os.path.dirname(__file__), "vue/vegalite.vue"))
    spec = traitlets.Dict().tag(sync=True)
    listen_to_click = traitlets.Bool(False).tag(sync=True)
    listen_to_hover = traitlets.Bool(False).tag(sync=True)
    on_click = traitlets.traitlets.Callable(None, allow_none=True)
    on_hover = traitlets.traitlets.Callable(None, allow_none=True)
    cdn = traitlets.Unicode(None, allow_none=True).tag(sync=True)

    def vue_altair_click(self, *args):
        if self.on_click is not None:
            self.on_click(*args)

    def vue_altair_hover(self, *args):
        if self.on_hover is not None:
            self.on_hover(*args)

    @traitlets.default("cdn")
    def _cdn(self):
        import solara.settings

        if not solara.settings.assets.proxy:
            return solara.settings.assets.cdn


class Navigator(v.VuetifyTemplate):
    template_file = os.path.realpath(os.path.join(os.path.dirname(__file__), "vue/navigator.vue"))
    location = traitlets.Unicode(None, allow_none=True).tag(sync=True)


class GridLayout(v.VuetifyTemplate):
    template_file = os.path.join(os.path.dirname(__file__), "vue/gridlayout.vue")
    gridlayout_loaded = traitlets.Bool(False).tag(sync=True)
    items = traitlets.Union([traitlets.List(), traitlets.Dict()], default_value=[]).tag(sync=True, **ipywidgets.widget_serialization)
    grid_layout = traitlets.List(default_value=cast(List[Dict], [])).tag(sync=True)
    draggable = traitlets.CBool(True).tag(sync=True)
    resizable = traitlets.CBool(True).tag(sync=True)
    cdn = traitlets.Unicode(None, allow_none=True).tag(sync=True)
    on_layout_updated = traitlets.traitlets.Callable(None, allow_none=True)
    col_num = traitlets.Int(12).tag(sync=True)
    row_height = traitlets.Int(30).tag(sync=True)

    def vue_layout_updated(self, *args):
        if self.on_layout_updated is not None:
            self.on_layout_updated(*args)

    @traitlets.default("cdn")
    def _cdn(self):
        import solara.settings

        if not solara.settings.assets.proxy:
            return solara.settings.assets.cdn


class HTML(v.VuetifyTemplate):
    template_file = os.path.realpath(os.path.join(os.path.dirname(__file__), "vue/html.vue"))
    tag = traitlets.Unicode("div").tag(sync=True)
    attributes = traitlets.Dict().tag(sync=True)
    unsafe_innerHTML = traitlets.Unicode(None, allow_none=True).tag(sync=True)


def watch():
    import ipyvue

    ipyvue.watch(os.path.realpath(os.path.dirname(__file__) + "/vue"))



import solara
from solara.website.components import CategoryLayout, Gallery


@solara.component
def Page(route_external=None):
    Gallery(route_external)


@solara.component
def Layout(children=[]):
    CategoryLayout(children=children)



import dataclasses
import io
import json
import logging
import os

# import tempfile
import threading
import time
import urllib.request
import uuid
from typing import IO, Any, Callable, Tuple, TypeVar, Union, cast

import solara
from solara.datatypes import FileContentResult, Result

logger = logging.getLogger("react-ipywidgets.extra.hooks")
chunk_size_default = 1024**2

__all__ = [
    "use_download",
    "use_fetch",
    "use_json_load",
    "use_json",
    "use_file_content",
    "use_uuid4",
    "use_unique_key",
    "use_state_or_update",
    "use_previous",
    "use_trait_observe",
]
T = TypeVar("T")
U = TypeVar("U")

MaybeResult = Union[T, Result[T]]


def use_retry(*actions: Callable[[], Any]):
    counter, set_counter = solara.use_state(0)

    def retry():
        for action in actions:
            action()
        set_counter(lambda counter: counter + 1)

    return counter, retry


def use_download(
    f: MaybeResult[Union[str, os.PathLike, IO]], url, expected_size=None, delay=None, return_content=False, chunk_size=chunk_size_default
) -> Result:
    from .use_thread import use_thread

    if not isinstance(f, Result):
        f = Result(value=f)
    assert isinstance(f, Result)
    content_length, set_content_length = solara.use_state(expected_size, key="content_length")
    downloaded_length = 0
    file_object = hasattr(f.value, "tell")
    if not file_object:
        file_path = cast(Union[str, os.PathLike], f.value)
        if os.path.exists(file_path) and expected_size is not None:
            file_size = os.path.getsize(file_path)
            if file_size == expected_size:
                downloaded_length = file_size
    downloaded_length, set_downloaded_length = solara.use_state(downloaded_length, key="downloaded_length")

    def download(cancel: threading.Event):
        assert isinstance(f, Result)
        nonlocal downloaded_length
        if expected_size is not None and downloaded_length == expected_size:
            return  # we already downloaded, but hooks cannot be conditional

        context: Any = None
        if file_object:
            context = solara.util.nullcontext()
            output_file = cast(IO, f.value)
        else:
            # f = cast(Result[Union[str, os.PathLike]], f)
            output_file = context = open(f.value, "wb")  # type: ignore

        with context:
            with urllib.request.urlopen(url) as response:
                content_length = int(response.info()["Content-Length"])
                logger.info("content_length for %r = %r", url, content_length)
                set_content_length(content_length)
                bytes_read = 0
                while not cancel.is_set():
                    chunk = response.read(chunk_size)

                    if delay:
                        time.sleep(delay)
                    if not chunk:
                        break
                    bytes_read += len(chunk)
                    output_file.write(chunk)
                    set_downloaded_length(bytes_read)
        return bytes_read

    result: Result[Any] = use_thread(download, [f, url])
    # maybe we wanna check this
    # download_is_done = downloaded_length == content_length

    if content_length is not None:
        progress = downloaded_length / content_length
    else:
        progress = 0

    return dataclasses.replace(result, progress=progress)


def use_fetch(url, chunk_size=chunk_size_default):
    # reuse the same file like object
    f = solara.use_memo(io.BytesIO, [url])
    result = use_download(f, url, return_content=True, chunk_size=chunk_size)
    return dataclasses.replace(result, value=f.getvalue() if result.progress == 1 else None)


def compose_result(head, *tail):
    return head


def ensure_result(input: MaybeResult[T]) -> Result[T]:
    if isinstance(input, Result):
        return input
    else:
        return Result(value=input)


def make_use_thread(f: Callable[[T], U]):
    from .use_thread import use_thread

    def use_result(input: MaybeResult[T]) -> Result[U]:
        input_result = ensure_result(input)

        def in_thread(cancel: threading.Event):
            if input_result.value:
                return f(input_result.value)

        return use_thread(in_thread, dependencies=[input_result.value])

    return use_result


@make_use_thread
def use_json_load(value: bytes):
    return json.loads(value)


def use_json(path):
    return use_fetch(path) | use_json_load


def use_file_content(path, watch=False) -> FileContentResult[bytes]:
    counter, retry = use_retry()

    def read_file(*ignore):
        try:
            with open(path, "rb") as f:
                return f.read()
        except Exception as e:
            return e

    result = None
    try:
        mtime = os.path.getmtime(path)
    except Exception:
        mtime = None

    content = solara.use_memo(read_file, dependencies=[path, mtime, counter])
    if result is not None:
        return result
    if isinstance(content, Exception):
        return FileContentResult[bytes](error=content, _retry=retry)
    else:
        return FileContentResult[bytes](value=content, _retry=retry)


def use_force_update() -> Callable[[], None]:
    """Returns a callable that can be used to force an update of a component.

    This is used when external state has change, and we need to re-render out component.
    """
    _counter, set_counter = solara.use_state(0, "force update counter")

    def updater():
        set_counter(lambda count: count + 1)

    return updater


def use_uuid4(dependencies=[]) -> str:
    """Generate a unique string using the uuid4 algorithm. Will only change when the dependencies change."""

    def make_uuid(*_ignore):
        return str(uuid.uuid4())

    return solara.use_memo(make_uuid, dependencies)


def use_unique_key(key: str = None, prefix: str = "", dependencies=[]) -> str:
    """Generate a unique string, or use key when not None. Dependencies are forwarded to `use_uuid4`."""
    uuid = use_uuid4(dependencies=dependencies)
    return prefix + (key or uuid)


def use_state_or_update(
    initial_or_updated: T, key: str = None, eq: Callable[[Any, Any], bool] = None
) -> Tuple[T, Callable[[Union[T, Callable[[T], T]]], None]]:
    """This is useful for situations where a prop can change from a parent
    component, which should be respected, and otherwise the internal
    state should be kept.
    """
    value, set_value = solara.use_state(initial_or_updated, key=key, eq=eq)

    def possibly_update():
        nonlocal value
        # only gets called when initial_or_updated changes
        set_value(initial_or_updated)
        # this make sure the return value gets updated directly
        value = initial_or_updated

    solara.use_memo(possibly_update, [initial_or_updated])
    return value, set_value


def use_previous(value: T, condition=True) -> T:
    ref = solara.use_ref(value)

    def assign():
        if condition:
            ref.current = value

    solara.use_effect(assign, [value])
    return ref.current


def use_trait_observe(has_trait_object, name):
    """Observe a trait on an object, and return its value.

    This is useful when you want your component to be in sync with a trait
    of a widget or [HasTraits object](https://traitlets.readthedocs.io/en/stable/).

    When the trait changes, your component will be re-rendered.

    See [use_dark_effective](/api/use_dark_effective) for an example.
    """
    counter = solara.use_reactive(0)
    counter.get()  # make the main component depend on this counter

    def connect():
        def update(change):
            counter.value += 1

        has_trait_object.observe(update, name)

        def cleanup():
            has_trait_object.unobserve(update, name)

        return cleanup

    solara.use_effect(connect, [has_trait_object, name])
    return getattr(has_trait_object, name)



import datetime as dt
from typing import Optional
from unittest.mock import MagicMock
import ipyvuetify as vw
import solara
from solara.lab.components.input_time import InputTime

now = dt.time(14, 30)
later = dt.time(16, 45)


def test_input_time_typing():
    def on_value(value: dt.time):
        pass

    InputTime(value=now, label="label", on_value=on_value)

    def on_value2(value: Optional[dt.time]):
        pass

    InputTime(value=now, label="label", on_value=on_value2)


def test_input_time():
    on_value = MagicMock()
    on_v_model = MagicMock()
    el = InputTime(value=now, label="label", on_value=on_value)
    box, rc = solara.render(el, handle_error=False)
    input = rc.find(vw.TextField)
    input.widget.observe(on_v_model, "v_model")

    input.widget.v_model = later.strftime("%H:%M")
    assert on_value.call_count == 1
    assert on_v_model.call_count == 1
    on_v_model.reset_mock()
    assert on_value.call_args[0][0] == later

    input.widget.v_model = ""
    assert on_value.call_count == 1
    assert on_v_model.call_count == 1
    assert on_value.call_args[0][0] == later
    assert input.widget.error
    assert input.widget.label == "label (Time cannot be empty)"

    input.widget.v_model = "10:17"
    assert on_value.call_count == 2
    assert on_v_model.call_count == 2
    assert on_value.call_args[0][0] == dt.time(10, 17)
    assert not input.widget.error
    assert input.widget.v_model == dt.time(10, 17).strftime("%H:%M")


def test_input_time_optional():
    on_value = MagicMock()
    on_v_model = MagicMock()
    el = InputTime(value=now, label="label", on_value=on_value, optional=True)
    box, rc = solara.render(el, handle_error=False)
    input = rc.find(vw.TextField)
    input.widget.observe(on_v_model, "v_model")

    input.widget.v_model = later.strftime("%H:%M")
    assert on_value.call_count == 1
    assert on_v_model.call_count == 1
    assert on_value.call_args[0][0] == later

    input.widget.v_model = ""
    assert on_value.call_count == 2
    assert on_v_model.call_count == 2
    assert on_value.call_args[0][0] is None
    assert not input.widget.error

    input.widget.v_model = "10:17"
    assert on_value.call_count == 3
    assert on_v_model.call_count == 3
    assert on_value.call_args[0][0] == dt.time(10, 17)
    assert not input.widget.error
    assert input.widget.v_model == dt.time(10, 17).strftime("%H:%M")


def test_input_time_incomplete_entry():
    on_value = MagicMock()
    on_v_model = MagicMock()

    @solara.component
    def Test():
        def update_value(value: dt.time):
            on_value(value)
            set_value(value)

        value, set_value = solara.use_state(now)

        InputTime(value=value, on_value=update_value, label="label")

    el = Test()
    box, rc = solara.render(el, handle_error=False)
    input = rc.find(vw.TextField)
    input.widget.observe(on_v_model, "v_model")

    input.widget.v_model = "10"
    assert on_value.call_count == 0
    assert on_v_model.call_count == 1
    assert on_value.call_args is None
    assert input.widget.error
    assert input.widget.v_model == "10"
    assert on_v_model.call_count == 1

    input.widget.v_model = "10:17"
    assert on_value.call_count == 1
    assert on_v_model.call_count == 2
    assert on_value.call_args is not None
    assert on_value.call_args[0][0] == dt.time(10, 17)
    assert not input.widget.error
    assert input.widget.v_model == dt.time(10, 17).strftime("%H:%M")


def test_input_time_invalid_format():
    on_value = MagicMock()
    on_v_model = MagicMock()

    @solara.component
    def Test():
        def update_value(value: dt.time):
            on_value(value)
            set_value(value)

        value, set_value = solara.use_state(now)

        InputTime(value=value, on_value=update_value, label="label", twelve_hour_clock=True)

    el = Test()
    box, rc = solara.render(el, handle_error=False)
    input = rc.find(vw.TextField)
    input.widget.observe(on_v_model, "v_model")

    input.widget.v_model = "1:30 PM"
    assert on_value.call_count == 1
    assert on_v_model.call_count == 1
    assert on_value.call_args is not None
    assert on_value.call_args[0][0] == dt.time(13, 30)
    assert not input.widget.error

    input.widget.v_model = "25:00 PM"
    assert on_value.call_count == 1
    assert on_v_model.call_count == 2  # This is still called to reflect change
    assert on_value.call_args[0][0] == dt.time(13, 30)  # Value remains unchanged due to invalid input
    assert input.widget.error
    assert input.widget.label == "label (Time 25:00 PM does not match format I:M p)"


def test_input_time_on_open_value():
    on_open_value = MagicMock()
    on_value = MagicMock()
    el = InputTime(value=now, label="label", on_value=on_value, on_open_value=on_open_value)
    box, rc = solara.render(el, handle_error=False)
    menu = rc.find(vw.VuetifyTemplate)
    assert menu is not None

    # Simulate opening the time picker
    menu.widget.show_menu = True
    assert on_open_value.call_count == 1
    assert on_open_value.call_args[0][0] is True

    # Simulate closing the time picker
    menu.widget.show_menu = False
    assert on_open_value.call_count == 2
    assert on_open_value.call_args[0][0] is False


def test_input_time_boundary_values():
    on_value = MagicMock()
    on_v_model = MagicMock()
    el = InputTime(value=now, label="label", on_value=on_value)
    box, rc = solara.render(el, handle_error=False)
    input = rc.find(vw.TextField)
    input.widget.observe(on_v_model, "v_model")

    # Test boundary value '00:00'
    input.widget.v_model = "00:00"
    assert on_value.call_count == 1
    assert on_v_model.call_count == 1
    assert on_value.call_args[0][0] == dt.time(0, 0)
    assert not input.widget.error

    # Test boundary value '23:59'
    input.widget.v_model = "23:59"
    assert on_value.call_count == 2
    assert on_v_model.call_count == 2
    assert on_value.call_args[0][0] == dt.time(23, 59)
    assert not input.widget.error

    # Test invalid boundary value '24:00'
    input.widget.v_model = "24:00"
    assert on_value.call_count == 2
    assert on_v_model.call_count == 3  # This is still called to reflect change
    assert on_value.call_args[0][0] == dt.time(23, 59)  # Value remains unchanged due to invalid input
    assert input.widget.error
    assert input.widget.label == "label (Time 24:00 does not match format H:M)"



"""
# use_router

"""

import solara
from solara.website.components import NoPage
from solara.website.utils import apidoc

title = "use_router"


Page = NoPage


__doc__ += apidoc(solara.use_router)  # type: ignore



from typing import List, Optional, Union

import reacton.ipyvuetify as v

import solara
from solara.util import _combine_classes


@solara.component
def Success(
    label: Optional[str] = None,
    icon: Union[bool, str, None] = True,
    dense=False,
    outlined=True,
    text=True,
    children=[],
    classes: List[str] = [],
    **kwargs,
):
    """Display a success message (green color).

    ## Arguments

     * `label`: the message to display
     * `icon`: if True, display a check icon, if False, don't display an icon, if a string,
               display the icon with that name ([Overview of available icons](https://pictogrammers.github.io/@mdi/font/4.9.95/)).
     * `dense`: if True, display the message in a dense format, using less vertical height.
     * `outlined`: if True (default), display the message in an outlined border, instead of a filled box.
     * `text`: if True (default), display the message in a text format, which applies a semi-transparent background.
     * `classes`: additional CSS classes to apply.
    """
    # vuetify doesn't accept True, but is ok with None for a default icon
    if icon is True:
        icon = None
    return v.Alert(
        type="success",
        text=text,
        outlined=outlined,
        dense=dense,
        icon=icon,
        children=[*([label] if label is not None else []), *children],
        class_=_combine_classes(classes),
        **kwargs,
    )


@solara.component
def Info(
    label: Optional[str] = None,
    icon: Union[bool, str, None] = True,
    dense=False,
    outlined=True,
    text=True,
    children=[],
    classes: List[str] = [],
    **kwargs,
):
    """Display a info message (blue color).

    ## Arguments

     * `label`: the message to display
     * `icon`: if True, display a info icon, if False, don't display an icon, if a string,
               display the icon with that name ([Overview of available icons](https://pictogrammers.github.io/@mdi/font/4.9.95/)).
     * `dense`: if True, display the message in a dense format, using less vertical height.
     * `outlined`: if True (default), display the message in an outlined border, instead of a filled box.
     * `text`: if True (default), display the message in a text format, which applies a semi-transparent background.
     * `classes`: additional CSS classes to apply.
    """
    if icon is True:
        icon = None
    return v.Alert(
        type="info",
        text=text,
        outlined=outlined,
        dense=dense,
        icon=icon,
        children=[*([label] if label is not None else []), *children],
        class_=_combine_classes(classes),
        **kwargs,
    )


@solara.component
def Warning(
    label: Optional[str] = None,
    icon: Union[bool, str, None] = True,
    dense=False,
    outlined=True,
    text=True,
    children=[],
    classes: List[str] = [],
    **kwargs,
):
    """Display a warning message (orange color).

    ## Arguments

     * `label`: the message to display
     * `icon`: if True, display a exclamation icon, if False, don't display an icon, if a string,
               display the icon with that name ([Overview of available icons](https://pictogrammers.github.io/@mdi/font/4.9.95/)).
     * `dense`: if True, display the message in a dense format, using less vertical height.
     * `outlined`: if True (default), display the message in an outlined border, instead of a filled box.
     * `text`: if True (default), display the message in a text format, which applies a semi-transparent background.
     * `classes`: additional CSS classes to apply.
    """
    if icon is True:
        icon = None
    return v.Alert(
        type="warning",
        text=text,
        outlined=outlined,
        dense=dense,
        icon=icon,
        children=[*([label] if label is not None else []), *children],
        class_=_combine_classes(classes),
        **kwargs,
    )


@solara.component
def Error(
    label: Optional[str] = None,
    icon: Union[bool, str, None] = True,
    dense=False,
    outlined=True,
    text=True,
    children=[],
    classes: List[str] = [],
    **kwargs,
):
    """Display an error message (red color).

    ## Arguments

     * `label`: the message to display
     * `icon`: if True, display a exclamation in a red triangle icon, if False, don't display an icon, if a string,
               display the icon with that name ([Overview of available icons](https://pictogrammers.github.io/@mdi/font/4.9.95/)).
     * `dense`: if True, display the message in a dense format, using less vertical height.
     * `outlined`: if True (default), display the message in an outlined border, instead of a filled box.
     * `text`: if True (default), display the message in a text format, which applies a semi-transparent background.
     * `classes`: additional CSS classes to apply.
    """
    if icon is True:
        icon = None
    return v.Alert(
        type="error",
        text=text,
        outlined=outlined,
        dense=dense,
        icon=icon,
        children=[*([label] if label is not None else []), *children],
        class_=_combine_classes(classes),
        **kwargs,
    )



import dataclasses
from enum import Enum
from pathlib import Path
from types import ModuleType
from typing import Any, Callable, Dict, Generic, List, Optional, TypeVar, Union

import reacton
from typing_extensions import Literal, TypedDict

T = TypeVar("T")
U = TypeVar("U")


@dataclasses.dataclass(frozen=True)
class Action:
    name: str
    icon: Optional[str] = None
    on_click: Optional[Callable] = None


@dataclasses.dataclass(frozen=True)
class ColumnAction(Action):
    on_click: Optional[Callable[[str], None]] = None


@dataclasses.dataclass(frozen=True)
class CellAction(Action):
    on_click: Optional[Callable[[str, int], None]] = None


def _fallback_retry():
    raise RuntimeError("Should not happen")


class ResultState(Enum):
    INITIAL = 1
    STARTING = 2
    WAITING = 3
    RUNNING = 4
    ERROR = 5
    FINISHED = 6
    CANCELLED = 7


@dataclasses.dataclass(frozen=True)
class Result(Generic[T]):
    value: Optional[T] = None
    error: Optional[Exception] = None
    state: ResultState = ResultState.INITIAL
    progress: Optional[float] = None

    def retry(self):
        # mypy does not like members that are callable
        # gets confused about self argument.
        # we wrap it to avoid hitting this error in user
        # code
        self._retry()  # type: ignore

    # can we avoid storing these into the dataclass?
    _retry: Callable[[], Any] = dataclasses.field(compare=False, default=lambda: None)
    cancel: Callable[[], Any] = dataclasses.field(compare=False, default=lambda: None)

    def __or__(self, next: Callable[["Result[T]"], "Result[U]"]):
        return next(self)


@dataclasses.dataclass(frozen=True)
class FileContentResult(Result[T]):
    @property
    def exists(self):
        return not isinstance(self.error, FileNotFoundError)


class AggregationCount(TypedDict):
    type: Literal["count"]


class AggregationSum(TypedDict):
    type: Literal["sum"]


JsonType = Union[None, int, str, bool, List[Any], Dict[str, Any]]

Aggregation = Union[AggregationCount, AggregationSum]


class PivotTableData(TypedDict):
    x: List[str]
    y: List[str]
    agg: str
    values: List[List[JsonType]]
    values_x: List[str]
    values_y: List[str]
    headers_x: List[List[str]]
    headers_y: List[List[str]]
    counts_x: int
    counts_y: int
    total: str


@dataclasses.dataclass(frozen=True)
class Route:
    """A route tells Solara which component to render for a given URL. (Not a Solara component!)

    ## Arguments

     * `path` - The path of the route. This is the part of the URL that you see in the browser.
     * `children` - (Optional) A list of child routes. These are routes that are nested under this route.
     * `module` - (Optional) The module that contains the component to render for this route, used for autorouting.
     * `component` - (Optional) The component to render for this route.
     * `layout` - (Optional) The layout to use for this route. If not specified, the default layout will be used.
     * `data` - (Optional) The data to pass to the component for this route, usage is up to the user.
     * `label` - (Optional) The label to use for this route, can be used for labeling tabs or links.

    ## See also

     * [Multipage](/documentation/advanced/howto/multipage).
     * [Understanding Routing](/documentation/advanced/understanding/routing).

    """

    path: str
    children: List["Route"] = dataclasses.field(default_factory=list)

    # these are free to use, depending on the implementation
    # see autorouting.py for how Solara uses them
    module: Optional[ModuleType] = None

    # in the autorouting implementation, this is the
    # the same as module.Page (unless we are rendering a markdown)
    component: Union[None, Callable, reacton.core.Component] = None
    layout: Union[None, Callable, reacton.core.Component] = None

    # in the autorouting implementation, this is the
    # path of the markdown file
    data: Any = None

    # Can be used for a title and/or a tab label
    label: Optional[str] = None

    # file corresponding to this route, can be used for
    # checking of content has changed (using mtime)
    file: Optional[Path] = None



"""This Page components sets in the package root and takes two non-optional arguments,
meaning it will catch urls like /viz/scatter/titanic and pass two argument to the Page component,
so we can render content dynamically.
"""

from typing import Optional

import plotly.express as px
import solara

from ... import data


def title(type: str, name: str):
    return f"Solara viz view: {type} - {name}"


@solara.component
def Page(type: Optional[str] = None, name: Optional[str] = None, x: Optional[str] = None, y: Optional[str] = None):
    type, set_type = solara.use_state_or_update(type)
    name, set_name = solara.use_state_or_update(name)
    x, set_x = solara.use_state_or_update(x)
    y, set_y = solara.use_state_or_update(y)

    with solara.ColumnsResponsive(12) as main:
        if type is None:
            type = "scatter"
            set_type("scatter")
        with solara.Sidebar():
            with solara.Card("Viz configuration"):
                solara.Select(label="dataset", value=name, values=list(data.dfs), on_value=set_name)
                solara.ToggleButtonsSingle(value=type, values=["scatter", "histogram"], on_value=set_type)
        if name not in data.dfs:
            set_name(list(data.dfs)[0])
        if name in data.dfs:
            df = data.dfs[name].df
            column_names = df.get_column_names()
            df = df.to_pandas_df()
            if x not in column_names:
                set_x(column_names[0])
            if y not in column_names:
                set_y(column_names[1])
            if x not in column_names:
                set_x(column_names[0])
            if y not in column_names:
                set_y(column_names[1])
            solara.Title(f"Solara demo » viz » {type} » {name}")
            fig = None
            if type == "scatter":
                with solara.Sidebar():
                    with solara.Card("Columns"):
                        solara.Select(label="x", value=x, values=column_names, on_value=set_x)
                        solara.Select(label="y", value=y, values=column_names, on_value=set_y)
                if x and y and x in column_names and y in column_names:
                    fig = px.scatter(df, x=x, y=y)
                else:
                    solara.Warning("Please provide x and y")
            elif type == "histogram":
                with solara.Sidebar():
                    with solara.Card("Columns"):
                        solara.Select(label="x", value=x, values=column_names, on_value=set_x)
                if x and x in column_names:
                    fig = px.histogram(df, x=x)
                else:
                    solara.Warning("Please provide x")
            else:
                solara.Error("Uknonwn ")
            if fig:
                solara.FigurePlotly(fig, dependencies=[name, type, x, y])
    return main



import shutil
from pathlib import Path

import pytest

import solara.lab
import solara.lifecycle
from solara.server import reload
from solara.server.app import AppScript

HERE = Path(__file__).parent

kernel_start_path = HERE / "solara_test_apps" / "kernel_start.py"


@pytest.mark.parametrize("as_module", [False, True])
def test_script_reload_component(tmpdir, kernel_context, extra_include_path, no_kernel_context, as_module):
    target = Path(tmpdir) / "kernel_start.py"
    shutil.copy(kernel_start_path, target)
    with extra_include_path(str(tmpdir)):
        on_kernel_start_callbacks = solara.lifecycle._on_kernel_start_callbacks.copy()
        callbacks_start = [k.callback for k in solara.lifecycle._on_kernel_start_callbacks]
        if as_module:
            app = AppScript(f"{target.stem}")
        else:
            app = AppScript(f"{target}")
        app.init()
        try:
            app.run()
            callback = app.routes[0].module.test_callback  # type: ignore
            callbacks = [k.callback for k in solara.lifecycle._on_kernel_start_callbacks]
            assert callbacks == [*callbacks_start, callback]
            prev = callbacks.copy()
            reload.reloader.reload_event_next.clear()
            target.touch()
            # wait for the event to trigger
            reload.reloader.reload_event_next.wait()
            app.run()
            callback = app.routes[0].module.test_callback  # type: ignore
            callbacks = [k[0] for k in solara.lifecycle._on_kernel_start_callbacks]
            assert callbacks != prev
            assert callbacks == [*callbacks_start, callback]
        finally:
            app.close()
            solara.lifecycle._on_kernel_start_callbacks.clear()
            solara.lifecycle._on_kernel_start_callbacks.extend(on_kernel_start_callbacks)


def test_on_kernel_start_cleanup(kernel_context, no_kernel_context):
    def test_callback_cleanup():
        pass

    cleanup = solara.lab.on_kernel_start(test_callback_cleanup)
    assert test_callback_cleanup in [k.callback for k in solara.lifecycle._on_kernel_start_callbacks]
    cleanup()
    assert test_callback_cleanup not in [k.callback for k in solara.lifecycle._on_kernel_start_callbacks]



from unittest.mock import MagicMock

import ipyvuetify as vw

import solara
from solara.lab.components.confirmation_dialog import ConfirmationDialog


def test_confirmation_dialog_ok():
    is_open = solara.reactive(True)
    on_ok = MagicMock()
    on_close = MagicMock()
    el = ConfirmationDialog(is_open, on_ok=on_ok, on_close=on_close, content="Hello")
    _, rc = solara.render(el, handle_error=False)
    buttons = rc.find(vw.Btn)
    assert len(buttons) == 2
    buttons[1].widget.click()
    assert on_ok.call_count == 1  # was OK button clicked?
    assert on_close.call_count == 1  # always triggered
    assert not is_open.value  # is dialog closed?


def test_confirmation_dialog_cancel():
    is_open = solara.reactive(True)
    on_ok = MagicMock()
    on_close = MagicMock()
    el = ConfirmationDialog(is_open, on_ok=on_ok, on_close=on_close, content="Hello")
    _, rc = solara.render(el, handle_error=False)
    buttons = rc.find(vw.Btn)
    assert len(buttons) == 2
    buttons[0].widget.click()
    assert on_ok.call_count == 0  # on_ok action should not have been executed
    assert on_close.call_count == 1  # always triggered
    assert not is_open.value  # is dialog closed?


def test_confirm_external_close():
    # e.g. when persistent=False, clicking away from the dialog closes it
    is_open = solara.reactive(True)
    on_ok = MagicMock()
    on_cancel = MagicMock()
    on_close = MagicMock()
    el = ConfirmationDialog(is_open, on_ok=on_ok, on_cancel=on_cancel, on_close=on_close, content="Hello")
    _, rc = solara.render(el, handle_error=False)
    dialog = rc.find(vw.Dialog)[0].widget
    assert dialog.v_model
    dialog.v_model = False  # trigger an external close, like escape or clicking away
    assert not is_open.value  # is dialog closed?
    assert on_ok.call_count == 0  # on_ok action should not have been executed
    assert on_cancel.call_count == 1  # on_cancel action should not have been executed
    assert on_close.call_count == 1  # always triggered


def test_confirmation_dialog_custom_button_no_onclick():
    is_open = solara.reactive(True)
    on_ok = MagicMock()
    my_button = solara.Button(label="Not OK")  # no on_click
    el = ConfirmationDialog(is_open, on_ok=on_ok, ok=my_button, content="Are you sure?")
    _, rc = solara.render(el, handle_error=False)
    buttons = rc.find(vw.Btn)
    assert len(buttons) == 2
    assert buttons[0].widget.children == ["Cancel"]
    assert buttons[1].widget.children == ["Not OK"]
    buttons[1].widget.click()
    assert on_ok.call_count == 1  # should still be called


def test_confirmation_dialog_custom_button_with_onclick():
    is_open = solara.reactive(True)
    values = []

    def on_ok():
        values.append("on_ok")

    def on_cancel():
        values.append("on_cancel")

    def on_click_ok():
        values.append("on_click_ok")

    def on_click_cancel():
        values.append("on_click_cancel")

    ok = solara.Button(label="Not OK", on_click=on_click_ok)
    cancel = solara.Button(label="Not Cancel", on_click=on_click_cancel)
    el = ConfirmationDialog(is_open, on_ok=on_ok, on_cancel=on_cancel, ok=ok, cancel=cancel, content="Are you sure?")
    _, rc = solara.render(el, handle_error=False)
    buttons = rc.find(vw.Btn)
    assert len(buttons) == 2
    assert buttons[1].widget.children == ["Not OK"]
    assert buttons[0].widget.children == ["Not Cancel"]
    buttons[1].widget.click()
    assert values == ["on_click_ok", "on_ok"]  # assert on_ok and on_click were both called, in that order
    values.clear()
    # now the same for cancel
    buttons[0].widget.click()
    assert values == ["on_click_cancel", "on_cancel"]



import solara
from solara.alias import rv


@solara.component
def Details(summary="Summary", children=[], expand=False):
    """Creates an expandable/collapsible section with a summary and additional children content

    ```solara
    import solara


    show_message = solara.reactive(True)
    disable = solara.reactive(False)


    @solara.component
    def Page():
        summary_text = "Click to expand for more details"
        additional_content = [
            "Additional detail 1",
            "Additional detail 2",
            "Additional detail 3"
        ]

        solara.Details(
            summary=summary_text,
            children=additional_content,
            expand=False
        )

    ```


    ## Arguments:

    * summary: String showing the summary text for the expandable section: Defaults "Summary"
    * children: List showing the children content of the expandable section: Defaults to an Empty list
    * expand: Boolean showing if the section is expanded or collapsed: Defaults to False
    """

    expand, set_expand = solara.use_state_or_update(expand)

    def on_v_model(v_model):
        if v_model is None:  # collapsed:
            set_expand(False)
        elif v_model == 0:
            set_expand(True)
        else:
            raise RuntimeError(f"v_model has odd value: {v_model}")

    with rv.ExpansionPanels(v_model=0 if expand else None, on_v_model=on_v_model) as main:
        with rv.ExpansionPanel():
            rv.ExpansionPanelHeader(children=[summary])
            rv.ExpansionPanelContent(children=children)
    return main



import solara


@solara.component
def Page1():
    solara.Button("hi1")


@solara.component
def Page2():
    solara.Button("hi2")


routes = [
    solara.Route(path="/", component=Page1, label="Hi1"),
    solara.Route(path="page2", component=Page2, label="Hi2"),
]



# this module will be conditionally imported in patch.py for now
# in the future, we may want to move the esm features of ipyreact
# into a separate package, and then we can import it unconditionally
import logging
import threading
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Tuple, Union

import ipyreact.importmap
import ipyreact.module

from solara.server import kernel_context

logger = logging.getLogger("solara.server.esm")
lock = threading.Lock()

_modules: Dict[str, Tuple[Union[str, Path], List[str]]] = {}
_modules_added_per_kernel: Dict[str, Dict[str, ipyreact.module.Module]] = defaultdict(dict)
_import_map_per_kernel: Dict[str, ipyreact.importmap.ImportMap] = {}


# in solara server, we'll monkey patch ipyreact.module with this
def define_module(name, module: Union[str, Path]):
    # collect the dependencies at this moment
    dependencies = list(_modules.keys())
    logger.info("define module %s %s (dependencies=%r)", name, module, dependencies)
    if name in _modules:
        old_module, dependencies = _modules[name]
    _modules[name] = (module, dependencies)
    if kernel_context.has_current_context():
        create_modules()
    return None


def get_module_names():
    return list(_modules.keys())


def create_modules():
    kernel_id = kernel_context.get_current_context().id
    _modules_added = _modules_added_per_kernel[kernel_id]
    logger.info("create modules %s", _modules)
    widgets = {}
    with lock:
        for name, (module, dependencies) in _modules.items():
            if name not in _modules_added:
                _modules_added[name] = create_module(name, module, dependencies=dependencies)
                logger.info("create module %s %s %s", name, module, dependencies)
            else:
                _modules_added[name].code = module if not isinstance(module, Path) else module.read_text(encoding="utf8")
                _modules_added[name].dependencies = dependencies
                logger.info("update module %s %s %s", name, module, dependencies)
            widgets[name] = _modules_added[name]
    return widgets


def create_module(name, module: Union[str, Path], dependencies: List[str]):
    return ipyreact.module.Module(code=module if not isinstance(module, Path) else module.read_text(encoding="utf8"), name=name, dependencies=dependencies)


def create_import_map():
    kernel_id = kernel_context.get_current_context().id
    with lock:
        if kernel_id not in _import_map_per_kernel:
            _import_map_per_kernel[kernel_id] = ipyreact.importmap.ImportMap(import_map=ipyreact.importmap._effective_import_map)
            logger.info("create import map %s", ipyreact.importmap._effective_import_map)
        else:
            _import_map_per_kernel[kernel_id].import_map = ipyreact.importmap._effective_import_map
            logger.info("update import map %s", ipyreact.importmap._effective_import_map)
    return
